#include "predef.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <startnet.h>
#include <udp.h>
#include <tcp.h>
#include <time.h>
#include <basictypes.h>
#include <multicast.h>
#include <ucos.h>
#include <bsp.h>
#include <math.h>	//fabs()
#include "ffs.h"
//
#include "ca_proto.h"
#include "caerr.h"

#include <effs_fat/fat.h>
#ifdef USE_MMC
#include <effs_fat/mmc_mcf.h>
#endif
#ifdef USE_CFC
#include <effs_fat/cfc_mcf.h>
#endif
//
#define REQ_BUF_SIZE	416	 	//request command data length, 32x13=416, search
#define MAX_REQ_SIZE	8192	//max request command packet size
#define RESP_BUF_SIZE	1400 	//response data length maximum, 104x13=1352
//
// several PV's might read/write at the same TCP channel 
// nType: PV_TYPE_DOUBLE, PV_TYPE_STRING
// wType: DBR_DOUBLE, DBR_STRING, DBR_TIME_DOUBLE,...
// nPvNdx: 0-based index for pvDouble[], pvString[]
// nPvSetNdx: 0-based index for pvSet[]
// nFdNdx: 0-based index for fdSet[]
//
UDP_CLIENT	UdpClient;
extern int nTxInterface; //UDP or RS232
extern NV_SETTINGS NV_Settings;

CAFD_SET fdSet[NFDS];		//nFdNdx, file descriptor set
PV_SET pvSet[NPVS];		//nPvSetNdx, PV set
//
// Declare a task stack for the UDP Reader task
DWORD UdpListenStack[USER_TASK_STK_SIZE];
DWORD UdpListen2Stack[USER_TASK_STK_SIZE];
DWORD TcpListenStack[USER_TASK_STK_SIZE];
DWORD ServerBeaconStack[USER_TASK_STK_SIZE];
DWORD PVCheckStack[USER_TASK_STK_SIZE];
//
void UdpListenTask( void *pd );
void UdpListen2Task( void *pd );
void ServerBeaconTask(void *pd);
void TcpListenTask(void *pd);
void PVCheckTask(void *pd);

extern BOOL bSendBeacon;
//
void CAS_Init(void)
{
	int i;
	//print information
	printf("CAS Initialization process\r\n");
	CheckNVSettings();
	//
	if (OSTaskCreate( UdpListenTask,
			( void * )CA_SERVER_PORT,	//5064
			&UdpListenStack[USER_TASK_STK_SIZE],
			UdpListenStack,
			UDP_LISTEN_PRIO ) != OS_NO_ERR)
			iprintf("UdpListenTask creation error\n");
	if (OSTaskCreate( UdpListen2Task,
			( void * )UDP_SERVER_PORT,	//5064
			&UdpListen2Stack[USER_TASK_STK_SIZE],
			UdpListen2Stack,
			UDP_SERVER_PRIO ) != OS_NO_ERR)
			iprintf("UdpListen2Task creation error\n");
	if (OSTaskCreate( TcpListenTask,
			( void * )CA_SERVER_PORT,
			&TcpListenStack[USER_TASK_STK_SIZE],
			TcpListenStack,
			TCP_LISTEN_PRIO ) != OS_NO_ERR)
			iprintf("TcpListenTask creation error\n");
	//nPortNum = CA_REPEATER_PORT;
	if (OSTaskCreate( ServerBeaconTask,
			( void * )CA_REPEATER_PORT,		//5065
			&ServerBeaconStack[USER_TASK_STK_SIZE],
			ServerBeaconStack,
			BEACON_PRIO ) != OS_NO_ERR)
			iprintf("ServerBeaconTask creation error\n");
	if (OSTaskCreate( PVCheckTask,
			( void * )CA_REPEATER_PORT,
			&PVCheckStack[USER_TASK_STK_SIZE],
			PVCheckStack,
			PV_CHECK_PRIO ) != OS_NO_ERR)
			iprintf("PVCheckTask creation error\n");
	
	for (i = 0; i < NFDS; i++)
		memset((BYTE *)&fdSet[i], 0, sizeof(CAFD_SET));
	for (i = 0; i < NPVS; i++)
		PVS_Clear(i);
	//
	IOC_Init();
}

DWORD GetNewSID()
{
	//return 1;
	static DWORD dwSID = 0;
	dwSID++;
	return dwSID;
}
//
// response to EVENT_ADD: DO_UPDATE
// response to camonitor: DONT_UPDATE
int MakePacket_EventAdd(int nPvSetNdx, BOOL bUpdate, CA_PACKET *pcaPkt)
{
	int nPvNdx;
	int nSize;
	if (bUpdate == DO_UPDATE)
	{
		nPvNdx = pvSet[nPvSetNdx].nPvNdx;
		IOC_GetData(nPvNdx, pvSet[nPvSetNdx].nType); //get physical value --> xNewvalue
		PV_UpdateData(pvSet[nPvSetNdx].nType);		//xNewValue --> xValue
	}
	pcaPkt->caHeader.wCommand = CA_PROTO_EVENT_ADD;
	pcaPkt->caHeader.wDataType = pvSet[nPvSetNdx].wEventDataType;
	pcaPkt->caHeader.wDataCount = 1;
	pcaPkt->caHeader.dwParam1 = ECA_NORMAL;
	pcaPkt->caHeader.dwParam2 = pvSet[nPvSetNdx].dwCSID;
	//wEventDataType = DBR_TIME_DOUBLE, DBR_TIME_STRING
	pcaPkt->caHeader.wPayloadSize = PV_GetPayload(pvSet[nPvSetNdx].wEventDataType, 
		pvSet[nPvSetNdx].nPvNdx, pcaPkt->bData);
	nSize = CA_HEADER_SIZE + (int)pcaPkt->caHeader.wPayloadSize;
	return nSize;
}
/*
	called by UDP or TCP
 */
int GetResponse(int nProtocol, BYTE *pRequest, BYTE *pResponse, int nLen)
{
	int nDataPtr, nSize;
	int nRespDataSize;
	CA_HEADER caRequest;
	CA_PACKET caResponsePkt;
	BYTE bResponse;
	BYTE bMatchPV = 0;
	int nType, nPvNdx;
	//
	nRespDataSize = 0;
	for (nDataPtr = 0; nDataPtr < nLen; )
	{
		bResponse = 1;
		memmove((BYTE *)&caRequest, pRequest, CA_HEADER_SIZE);
		memmove(&(caResponsePkt.caHeader), &caRequest, CA_HEADER_SIZE);
		switch (caRequest.wCommand)
		{
			case CA_PROTO_VERSION:
				caResponsePkt.caHeader.wPayloadSize = 0;
				caResponsePkt.caHeader.wDataType = 0;	//priority
				caResponsePkt.caHeader.wDataCount = CA_PROTO_MIN_VER;
				bResponse = (nProtocol == PROTO_UDP) ? 1 : 0;
				break;
			case CA_PROTO_SEARCH:
				//caRequest.dwParam1 = channel ID
				//caRequest.dwParam2 = channel ID
				//if (caRequest.wDataType == DO_REPLY)
				//
				caResponsePkt.caHeader.wPayloadSize = 8;
				//TCP Port number of server that responded.
				caResponsePkt.caHeader.wDataType = CA_SERVER_PORT;
				caResponsePkt.caHeader.wDataCount = 0;
				caResponsePkt.caHeader.dwParam1 = 0xFFFFFFFF;	//temporary SID
				caResponsePkt.caHeader.dwParam2 = caRequest.dwParam2;	//channel CID
				caResponsePkt.bData[0] = 0;
				caResponsePkt.bData[1] = CA_PROTO_MIN_VER;
				if (PV_IsExist((char *)&pRequest[CA_HEADER_SIZE], &nType, &nPvNdx) == FALSE)
					bResponse = 0;	//don't response if PV is not matched
				else 
					bMatchPV++;
				break;
			default:
				bResponse = 0;
				break;
		}
		nSize = CA_HEADER_SIZE + (int)caRequest.wPayloadSize;
		pRequest += nSize;
		nDataPtr = nDataPtr + nSize;
		if (bResponse == 1)
		{
			nSize = CA_HEADER_SIZE + (int)caResponsePkt.caHeader.wPayloadSize;
			memmove(pResponse, (BYTE *)&caResponsePkt, nSize);
			pResponse += nSize;
			nRespDataSize += nSize;
		}
		if (nRespDataSize > RESP_BUF_SIZE - 32)	//response packet size has reach limit
			break;
	}
	// UDP: VERSION, SEARCH, SEARCH, SEARCH,...
	// TCP: VERSION, CLIENT_NAME, HOST_NAME, CREATE)CHANNEL (bMatchPV might be 0)
	if (bMatchPV == 0 && (nProtocol == PROTO_UDP))	//no match PV found
		return 0;
	return nRespDataSize;
}
/* response to DBR_CTRL_DOUBLEonly */
int GetReadNotifyResponse(int nFdNdx, BYTE *pRequest, BYTE *pResponse, int nLen)
{
	int nDataPtr, nSize;
	int nRespDataSize;
	CA_HEADER caRequest;
	CA_PACKET caResponsePkt;
	//WORD wMask;
	BYTE bResponse;
	//int nType, nPvNdx; 
	int nPvNdx, nPvSetNdx;
	//char szHost[MAX_NAME_SIZE_P1], szClient[MAX_NAME_SIZE_P1];
	//
	nRespDataSize = 0;
	for (nDataPtr = 0; nDataPtr < nLen; )
	{
		bResponse = 0;
		memmove((BYTE *)&caRequest, pRequest, CA_HEADER_SIZE);
		memmove(&(caResponsePkt.caHeader), &caRequest, CA_HEADER_SIZE);
		switch (caRequest.wCommand)
		{
			case CA_PROTO_READ_NOTIFY:
				//caRequest.dwParam1 = temporary SID, no change
				//caRequest.dwParam2 = client provided IOID, no change
				// stringin or analog input
				nPvSetNdx = PVS_GetPvSetNdxFromSID(caRequest.dwParam1);
				if (nPvSetNdx < 0)	//invalid nPvSetNdx
					break;
				//read limitation, every one can read it?
				if ((pvSet[nPvSetNdx].nAuthorizedRW & CA_PROTO_ACCESS_RIGHT_READ) != CA_PROTO_ACCESS_RIGHT_READ)
					break;
				nPvNdx = pvSet[nPvSetNdx].nPvNdx; 
				pvSet[nPvSetNdx].dwIOID = caRequest.dwParam2;
				caResponsePkt.caHeader.wDataCount = 1;
				caResponsePkt.caHeader.wDataType = caRequest.wDataType;
				// READ_NOTIFY,EVENT_ADD,READ_NOTIFY,EVENT_ADD	--> #caget --> response EVENT_READ_NOTIFY only
				// READ_NOTIFY,READ_NOTIFY                      --> archive engine request -> response to READ_NOTIFY(DBR_CTRL_DOUBLE)
				if (caRequest.wDataType != DBR_CTRL_DOUBLE)
				{	//no response to DBR_CTRL_STRING
					bResponse = 0;
					break;
				}
				caResponsePkt.caHeader.dwParam1 = 1;	//always 1, SID 
				IOC_GetData(nPvNdx, pvSet[nPvSetNdx].nType); //get physical value --> xNewvalue
				PV_UpdateData(pvSet[nPvSetNdx].nType);		//xNewValue --> xValue, clear bChanged flag
				caResponsePkt.caHeader.wPayloadSize = PV_GetPayload(caRequest.wDataType, nPvNdx, caResponsePkt.bData);
				bResponse = 1;
				break;
			default:
				bResponse = 0;
				break;
		}
		nSize = CA_HEADER_SIZE + (int)caRequest.wPayloadSize;
		pRequest += nSize;
		nDataPtr = nDataPtr + CA_HEADER_SIZE + (int)caRequest.wPayloadSize;
		if (bResponse == 1)
		{
			nSize = CA_HEADER_SIZE + (int)caResponsePkt.caHeader.wPayloadSize;
			memmove(pResponse, (BYTE *)&caResponsePkt, nSize);
			pResponse += nSize;
			nRespDataSize += nSize;
		}
		if (nRespDataSize > RESP_BUF_SIZE - 32)	//response packet size has reach limit
			break;
	}
	return nRespDataSize;	//response data size
}

/*
 This function is called by TCP listener
 nFdNdx = 0 ~ (NFDS-1)
*/
int SetAction(int nFdNdx, BYTE *pRequest, BYTE *pResponse, int nLen)
{
	int nDataPtr, nSize;
	int nRespDataSize;
	CA_HEADER caRequest;
	CA_PACKET caResponsePkt;
	WORD wMask;
	BYTE bResponse;
	int nType, nPvNdx, nPvSetNdx;
	char szHost[MAX_NAME_SIZE_P1], szClient[MAX_NAME_SIZE_P1];
	//
	nRespDataSize = 0;
	for (nDataPtr = 0; nDataPtr < nLen; )
	{
		bResponse = 0;
		memmove((BYTE *)&caRequest, pRequest, CA_HEADER_SIZE);
		memmove(&(caResponsePkt.caHeader), &caRequest, CA_HEADER_SIZE);
		switch (caRequest.wCommand)
		{
			case CA_PROTO_ECHO:		//V4.39
				//memset(&(caResponsePkt.caHeader), 0, CA_HEADER_SIZE);
				caResponsePkt.caHeader.wCommand = CA_PROTO_ECHO;
				bResponse = 1;
				break;
			case CA_PROTO_WRITE:
				//caRequest.dwParam1 = temporary SID
				//caRequest.dwParam2 = client provided IOID
				//caRequest.wDataType == DBR_DOUBLE or DBR_STRING
				//stringout or analog output
				nPvSetNdx = PVS_GetPvSetNdxFromSID(caRequest.dwParam1);
				if ((pvSet[nPvSetNdx].nAuthorizedRW & CA_PROTO_ACCESS_RIGHT_WRITE) == CA_PROTO_ACCESS_RIGHT_WRITE)
					PVS_SetData(nPvSetNdx, caRequest.wDataType, (char *)&pRequest[CA_HEADER_SIZE]); //might be DOUBLE or STRING
				bResponse = 0;
				if ((PVS_GetMonitorMask(nPvSetNdx) & DBE_VALUE) != 0)
				{	//if this parameter is monitored at this pvSet[] channel
					MakePacket_EventAdd(nPvSetNdx, DO_UPDATE, &caResponsePkt);
					bResponse = 1;
				}
				break;
			case CA_PROTO_READ_NOTIFY:
				//caRequest.dwParam1 = temporary SID, no change
				//caRequest.dwParam2 = client provided IOID, no change
				// stringin or analog input
				nPvSetNdx = PVS_GetPvSetNdxFromSID(caRequest.dwParam1);
				if (nPvSetNdx < 0)	//invalid nPvSetNdx
					break;
				//read limitation, every one can read it?
				if ((pvSet[nPvSetNdx].nAuthorizedRW & CA_PROTO_ACCESS_RIGHT_READ) != CA_PROTO_ACCESS_RIGHT_READ)
					break;
				nPvNdx = pvSet[nPvSetNdx].nPvNdx; 
				pvSet[nPvSetNdx].dwIOID = caRequest.dwParam2;
				caResponsePkt.caHeader.wDataCount = 1;
				caResponsePkt.caHeader.wDataType = caRequest.wDataType;
				// READ_NOTIFY,EVENT_ADD,READ_NOTIFY,EVENT_ADD	--> #caget --> response EVENT_READ_NOTIFY only
				// READ_NOTIFY,READ_NOTIFY                      --> archive engine request -> response to READ_NOTIFY(DBR_CTRL_DOUBLE)
				if (caRequest.wDataType == DBR_CTRL_DOUBLE) //used with archive engine
				{
					bResponse = 0;
					break;
				}
				caResponsePkt.caHeader.dwParam1 = 1;	//always 1, SID 
				IOC_GetData(nPvNdx, pvSet[nPvSetNdx].nType); //get physical value --> xNewvalue
				PV_UpdateData(pvSet[nPvSetNdx].nType);		//xNewValue --> xValue, clear bChanged flag
				caResponsePkt.caHeader.wPayloadSize = PV_GetPayload(caRequest.wDataType, nPvNdx, caResponsePkt.bData);
				bResponse = (caResponsePkt.caHeader.wPayloadSize != 0) ? 1 : 0;
				break;
			case CA_PROTO_CLIENT_NAME:
				strncpy(szClient, (char *)&pRequest[CA_HEADER_SIZE], MAX_NAME_SIZE);
				szClient[MAX_NAME_SIZE] = 0;
				break;
			case CA_PROTO_HOST_NAME:
				strncpy(szHost, (char *)&pRequest[CA_HEADER_SIZE], MAX_NAME_SIZE);
				szHost[MAX_NAME_SIZE] = 0;
				break;
			case CA_PROTO_CREATE_CHAN:
				//caRequest.dwParam1 = client provided channel ID, 
				//caRequest.dwParam2 = minor version
				nPvSetNdx = PVS_GetEmptyPvSet();
				if (nPvSetNdx < 0)	//pvSet full
					break;
				strcpy(pvSet[nPvSetNdx].szPV, (char *)&pRequest[CA_HEADER_SIZE]);
				if (PV_IsExist(pvSet[nPvSetNdx].szPV, &nType, &nPvNdx) == FALSE)
					break;
				// check read/write authority
				pvSet[nPvSetNdx].nFdNdx = nFdNdx;
				pvSet[nPvSetNdx].nType = nType;
				pvSet[nPvSetNdx].wDataType = PV_GetDataType(nType); //DBR_TIME_DOUBLE or DBR_TIME_STRING
				pvSet[nPvSetNdx].dwCID = caRequest.dwParam1;
				pvSet[nPvSetNdx].dwSID = GetNewSID();	//get a new SID for new PV
				pvSet[nPvSetNdx].nPvNdx = nPvNdx;
				if (CheckAuthorization(fdSet[pvSet[nPvSetNdx].nFdNdx].iaClient, szHost, szClient, nPvSetNdx) < 0)
				{
					//pvSet[nPvSetNdx].nAuthorizedRW = CA_PROTO_ACCESS_RIGHT_READ; //default read only
					pvSet[nPvSetNdx].nAuthorizedRW = CA_PROTO_ACCESS_RIGHT_NONE;
					break;
				}
				if ((pvSet[nPvSetNdx].nAuthorizedRW & CA_PROTO_ACCESS_RIGHT_WRITE) == CA_PROTO_ACCESS_RIGHT_WRITE)
					pvSet[nPvSetNdx].nAuthorizedRW |= CA_PROTO_ACCESS_RIGHT_READ; //no write only
				
				PVS_SetMonitorMask(nPvSetNdx, 0);	//clear monitor mask flag
				//
				caResponsePkt.caHeader.wCommand = CA_PROTO_ACCESS_RIGHT;
				caResponsePkt.caHeader.wPayloadSize = 0;
				caResponsePkt.caHeader.wDataType = 0;	//priority
				caResponsePkt.caHeader.wDataCount = 0;  //spec = CA_PROTO_MIN_VER;
				//caResponsePkt.caHeader.dwParam2 = fdSet[nFdNdx].dwCID; //no change
				caResponsePkt.caHeader.dwParam2 = PV_GetAccessRight(nType, nPvNdx);
				nSize = CA_HEADER_SIZE;
				memmove(pResponse, (BYTE *)&caResponsePkt, nSize);
				pResponse += nSize;
				nRespDataSize += nSize;
				//
				caResponsePkt.caHeader.wCommand = CA_PROTO_CREATE_CHAN;
				caResponsePkt.caHeader.wPayloadSize = 0;
				//DBR_DOUBLE or DBR_STRING  
				caResponsePkt.caHeader.wDataType = pvSet[nPvSetNdx].wDataType;	//???
				caResponsePkt.caHeader.wDataCount = 1;
				caResponsePkt.caHeader.dwParam1 = pvSet[nPvSetNdx].dwCID;
				caResponsePkt.caHeader.dwParam2 = pvSet[nPvSetNdx].dwSID;	//temporary SID	
				bResponse = 1;
				break;
			case CA_PROTO_EVENT_ADD:	//camonitor
				//caRequest payload: LO(4), HI(4), TO(4), Mask(2), ZERO(2)
				//caRequest.dwParam1 = temporary SID
				//caRequest.dwParam2 = client provided subscription ID (CSID)
				nPvSetNdx = PVS_GetPvSetNdxFromSID(caRequest.dwParam1);
				if (nPvSetNdx < 0)	//pvSet not exist
					break;
				pvSet[nPvSetNdx].dwCSID = caRequest.dwParam2;
				pvSet[nPvSetNdx].wEventDataType = caRequest.wDataType; //DBR_TIME_DOUBLE, DBR_TIME_STRING
				wMask = ByteToWord(pRequest + CA_HEADER_SIZE + 12);
				PVS_SetMonitorMask(nPvSetNdx, wMask); //DBE_VALUE(0x01), DBE_ALARM(0x04), DBE_LOG(0x02)
				// make EVENT_ADD response packet
				MakePacket_EventAdd(nPvSetNdx, DO_UPDATE, &caResponsePkt);
				bResponse = 1;				
				break;
			/* case CA_PROTO_EVENT_CANCEL:	//opposite to EVENT_ADD
			 *	//caRequest.dwParam1 = temporary SID
				//caRequest.dwParam2 = client provided subscription ID (CSID)
			 * 	break;
			 * case CA_PROTO_CLEAR_CHANNEL:	//opposite to CREATE_CHAN
				//dwParam1: temporary SID
				//dwParam2: channel CID
				nPvSetNdx = PVS_GetPvSetNdxFromSID(caRequest.dwParam1);
				if (nPvSetNdx >= 0) //TBD
				{
					nFdNdx = pvSet[nPvSetNdx].nFdNdx;
					PVS_ClearFD(nFdNdx);
					if (fdSet[nFdNdx].fdConn) close(fdSet[nFdNdx].fdConn);
					fdSet[nFdNdx].fdConn = 0;
					memset(&fdSet[nFdNdx], 0, sizeof(ID_SET));
				}
				bResponse = 0;
			*/
			default:
				bResponse = 0;
				break;
		}
		nSize = CA_HEADER_SIZE + (int)caRequest.wPayloadSize;
		pRequest += nSize;
		nDataPtr = nDataPtr + CA_HEADER_SIZE + (int)caRequest.wPayloadSize;
		if (bResponse == 1)
		{
			nSize = CA_HEADER_SIZE + (int)caResponsePkt.caHeader.wPayloadSize;
			memmove(pResponse, (BYTE *)&caResponsePkt, nSize);
			pResponse += nSize;
			nRespDataSize += nSize;
		}
		if (nRespDataSize > RESP_BUF_SIZE - 32)	//response packet size has reach limit
			break;
	}
	return nRespDataSize;	//response data size
}
/*-------------------------------------------------------------------
 * This task will wait for incoming UDP packets and process them.
 * Listening at 5064
 -------------------------------------------------------------------*/
int GetRequestRemainByte(BYTE *bRequestBuf, int nRequestLen)
{
	BYTE *p1;
	CA_HEADER *p2;
	int nNdx = 0, nNextNdx;
	p1 = bRequestBuf;
	while (nNdx < nRequestLen)
	{
		p2 = (CA_HEADER *)p1;
		nNextNdx = nNdx + p2->wPayloadSize + CA_HEADER_SIZE;
		if (nNextNdx > MAX_REQ_SIZE)	//illegal value
			return -1;
		if (nNextNdx > nRequestLen)
			break;
		nNdx = nNextNdx;
		p1 = bRequestBuf + nNdx;
	}
	return (nRequestLen - nNdx); 
}
//
void UdpListenTask( void *pd )

{
	int nDestPort = ( int ) pd;
	IPADDR iaSource;
	IPADDR iaDestination;
	int nSrcPort;
	int nRequestLen, nResponseLen;
	int nReqRemainByte;
	//
	BYTE bRequestBuf[REQ_BUF_SIZE];
	BYTE bResponseBuf[RESP_BUF_SIZE];
	int nReqDataNdx, nReqDataSize;
	char *p;
	//
	nDestPort = CA_SERVER_PORT;
	iprintf( "UdpListenTask Listening on port %d\r\n", nDestPort );
	// Create FIFO to store incoming packets and initialize it
	
	OS_FIFO fifo;
	OSFifoInit( &fifo );
	// Register to listen for UDP packets on port number 'port'
	//!!!don't use RegisterMulticastFifo(), it is not working
	RegisterUDPFifo(nDestPort, &fifo);
	while ( 1 )
	{
	    // We construct a UDP packet object using the FIFO. 
		// This constructor will only return when we have received a packet
		UDPPacket upkt( &fifo,
		              0 /* Replace this 0 with a tick count to have a time out delay */ );
		// Did we get a valid packet or just time out?
		if ( upkt.Validate() )
		{
			nReqDataSize = (int)upkt.GetDataSize();
			iaSource = upkt.GetSourceAddress();
			iaDestination = upkt.GetDestinationAddress();
			nSrcPort = upkt.GetSourcePort();
			nDestPort = upkt.GetDestinationPort();
			//if (nDestPort == CA_SERVER_PORT) {
			p = (char *)malloc(nReqDataSize);
			memmove(p, upkt.GetDataBuffer(), nReqDataSize );
			nReqRemainByte = 0;
			
			for (nReqDataNdx = 0; nReqDataNdx < nReqDataSize; ) {
			    if (REQ_BUF_SIZE - nReqRemainByte > nReqDataSize - nReqDataNdx)
			        nRequestLen = nReqDataSize - nReqDataNdx;
			    else
			        nRequestLen = REQ_BUF_SIZE - nReqRemainByte;
			    if ( nRequestLen > 0 )
			    {
			        memmove(&bRequestBuf[nReqRemainByte], p + nReqDataNdx, nRequestLen );
			        nRequestLen = nRequestLen + nReqRemainByte;
			        nReqRemainByte = GetRequestRemainByte(bRequestBuf, nRequestLen);
			        if (nReqRemainByte < 0)
			            break;             
			        nResponseLen = GetResponse(PROTO_UDP, bRequestBuf, bResponseBuf, nRequestLen - nReqRemainByte);
		        	//
					//if ((GetAuthorizedIP(iaSource) >= 0) && (nLen > 0))
					if (nResponseLen > 0) //don't check IP here, might use host and client name
					{   //send response
						{
							UDPPacket UdpTxPkt;
							UdpTxPkt.SetSourcePort( nDestPort );
							UdpTxPkt.SetDestinationPort( nSrcPort );
							UdpTxPkt.AddData(bResponseBuf, nResponseLen);
							UdpTxPkt.Send(iaSource);
		            	}
					}
					if (nReqRemainByte > 0)
					    memmove(bRequestBuf, &bRequestBuf[nRequestLen - nReqRemainByte], nReqRemainByte);
				}
				else
					break;
				nReqDataNdx += nRequestLen;
			} //for (nReqDataNdx = 0;
			if (p) free(p);
			//}
			#if DEBUG_CA == 1
		    iprintf( "RxUDP %d bytes from :", nReqDataSize );
			ShowIP( iaSource );
			iprintf( ":%d to ", nSrcPort );
			ShowIP( iaDestination );
			iprintf( ":%d\r\n", nDestPort );
			#endif
		} //if ( upkt.Validate() )
	}   //end of while (1)
}
//
void UdpListen2Task( void *pd )
{
	IPADDR iaDestination;
	int nDestPort = ( int ) pd;
	//
	BYTE bRequestBuf[REQ_BUF_SIZE];
	int nReqLen;
	//
	nDestPort = UDP_SERVER_PORT;
	iprintf( "UdpListen2Task Listening on port %d\r\n", nDestPort );
	
	// Create FIFO to store incoming packets and initialize it
	OS_FIFO fifo;
	OSFifoInit( &fifo );
	
	// Register to listen for UDP packets on port number 'port'
	//!!!don't use RegisterMulticastFifo(), it is not working
	RegisterUDPFifo(nDestPort, &fifo);
	
	while ( 1 )
	{
		// We construct a UDP packet object using the FIFO. 
		// This constructor will only return when we have received a packet
		UDPPacket upkt( &fifo,
                      0 /* Replace this 0 with a tick count to have a time out delay */ );

		// Did we get a valid packet or just time out?
		if ( upkt.Validate() )
		{
			if (nDestPort == UDP_SERVER_PORT) {
				nReqLen = (int)upkt.GetDataSize();
				memmove( bRequestBuf, upkt.GetDataBuffer(), nReqLen );
				nTxInterface = INTERFACE_UDP;
				UdpClient.iaSource = upkt.GetSourceAddress();
				iaDestination = upkt.GetDestinationAddress();
				UdpClient.nSrcPort = upkt.GetSourcePort();
				UdpClient.nDestPort = upkt.GetDestinationPort();
				DecipherCommand((char *)bRequestBuf);
			}
			#if DEBUG_CA == 1
			iprintf( "Rx2UDP %d bytes from :", nReqLen );
			ShowIP( UdpClient.iaSource );
			iprintf( ":%d to ", UdpClient.nSrcPort );
			ShowIP( iaDestination );
			iprintf( ":%d\r\n", nDestPort );
			#endif
		}
	}	//end of while (1)
}
//
/*-------------------------------------------------------------------
 * This task will wait for incoming UDP packets and process them.
 * Listening at 5064
 -------------------------------------------------------------------*/
//
void ServerBeaconTask(void *pd)
{
	int nDestPort = ( int ) pd; //CA_SERVER_PORT
	CA_HEADER caResponse;
	IPADDR iaBeacon;	//boardcast destination
	int nTickInterval;
	OS_FIFO fifo;
	static DWORD dwBeaconID = 0;
	DWORD dwV;
	//
	nDestPort = CA_REPEATER_PORT; //5065 (R5)
	OSFifoInit( &fifo );
	#if DEBUG_CA == 1
	iprintf( "ServerBeaconTask Listening on port %d\r\n", nDestPort );
	#endif
	//
	//EthernetIP: local host IP
	/*IPtoStringWithMask(EthernetIP, szBuffer, bMask);
	//if EthernetIP = "172.16.1.101" then szBuffer = "172.16.255.255"
	iaBeacon = AsciiToIp( szBuffer ); */
	dwV = (DWORD)NV_Settings.iaNetMask;
	iaBeacon = (IPADDR)((DWORD)EthernetIP | (~dwV));
	//
	caResponse.wCommand = CA_PROTO_RSRV_IS_UP; 
	caResponse.wPayloadSize = 0;
	caResponse.wDataType = CA_SERVER_PORT;	//TCP server listening port
	caResponse.wDataCount = 0;
	caResponse.dwParam2 = (DWORD)EthernetIP; //dwHostIP;
	//
	dwBeaconID = 0;
	nTickInterval = 2;	//start from 100 msec interval
	while (1)
	{
		caResponse.dwParam1 = dwBeaconID;
		//*
		if (bSendBeacon == 1)
		{
			#if DEBUG_CA == 1
			printf( "%.2f, beacon %ld to IP: ", (float)TimeTick/TICKS_PER_SECOND, dwBeaconID);
			ShowIP(iaBeacon);
			printf("\n");
			#endif
			//UDP broadcast from 2943 to 5065
			{
				UDPPacket pkt;
				pkt.SetSourcePort( 2943 );
				pkt.SetDestinationPort( nDestPort );
				pkt.AddData( (BYTE *)&caResponse, CA_HEADER_SIZE);
				pkt.Send( iaBeacon );
			}
			//*/
			dwBeaconID++;
			if (nTickInterval < TICKS_PER_SECOND * 15 / 2)
				nTickInterval *= 2;
			else 
				nTickInterval = TICKS_PER_SECOND * 15;
		}
		if (dwBeaconID % 39 == 0) //compensate 1.5 msec delay
			OSTimeDly(nTickInterval - 1);
		else
			OSTimeDly(nTickInterval);
	}
}
/*
	multicast listening TCP
	Listen for incoming TCP connections. You only need to call 
	listen() one time.  
	- Any IP address
	- Port number = CA_SERVER_PORT
	- Queue up to NFDS incoming connection requests
*/	
void TcpListenTask(void *pd)
{
	//int port = ( int ) pd;
	int nDestPort;
	int fdListen; 
	BYTE bRequestBuf[REQ_BUF_SIZE];
	BYTE bResponseBuf[RESP_BUF_SIZE];
	int i, nRequestLen, nResponseLen;
	int nReqRemainByte;
	//
	nDestPort = CA_SERVER_PORT; //5064 (S4)
	fdListen = listen( 0, nDestPort, 5 ); //INADDR_ANY = 0
	iprintf( "TCPListenTask Listening on port %d\r\n", nDestPort );
	while ( 1 )
	{
		// Declare file descriptor sets for select() 
		fd_set read_fds;
		fd_set error_fds;
		
		// Init the fd sets
		FD_ZERO( &read_fds );
		FD_ZERO( &error_fds );
		
		// Configure the fd sets so select() knows what to process. In
		// this case any fd data to be read, or an error.
		for ( i = 0; i < NFDS; i++ )
		{
			if ( fdSet[i].fdConn )
			{
				FD_SET( fdSet[i].fdConn, &read_fds );
				FD_SET( fdSet[i].fdConn, &error_fds );
			}
		}
		
		// select() should also process the listen fd
		FD_SET( fdListen, &read_fds );
		FD_SET( fdListen, &error_fds );
		
		/*
		select() will block until any fd has data to be read, or 
		has an error. When select() returns, read_fds and/or
		error_fds variables will have been modified to reflect
		the events. 
		*/ 
		select( FD_SETSIZE, &read_fds, ( fd_set * )0, &error_fds, 0 ); //nfds,*read_fds,*write_fds,*error_fds,timeout
		
		// If the listen fd has a connection request, accept it.
		if ( FD_ISSET( fdListen, &read_fds ) )
		{
			IPADDR client_ip;
			WORD   client_port;
			
			int fda = accept( fdListen, &client_ip, &client_port, 0 );
			
			// If accept() succeeded, find an open fd array slot
			if ( fda > 0) // && IsAuthorizedIP(client_ip))
			{
				for ( i = 0; i < NFDS; i++ )
				{
					if ( fdSet[i].fdConn == 0 )
					{
						fdSet[i].fdConn = fda;
						fdSet[i].iaClient = client_ip;
						#if DEBUG_CA == 1
						iprintf( "Added connection on fd[%d] = %d, ", i, fda );
						ShowIP(client_ip);
						iprintf(":%d\n", client_port);
						#endif
						fda = 0;
						break;
					}
				}
			}
			// If no array positions are open, close the connection
			if ( fda )
			{
				close( fda );
				#if DEBUG_CA == 1
				iprintf("Server Full\r\n");
				#endif
			}
		}

		// If the listen fd has an error, close it and reopen
		if ( FD_ISSET( fdListen, &error_fds ) )
		{
			close( fdListen );
			fdListen = listen( 0, nDestPort, 5 ); //INADDR_ANY = 0
		}

		// Process each fd array element and check it against read_fds
		// and error_fds. 
		for ( i = 0; i < NFDS; i++ )
		{
			if ( fdSet[i].fdConn == 0)
				continue;
			// Check for data to be read
			if ( FD_ISSET( fdSet[i].fdConn, &read_fds ) )
			{	
				//PROTO_VERSION,CLIENT_NAME,HOST_NAME,CREATE_CHAN,CREATE_CHAN,... (16+24+40)+32xN=464(if N=12)
				//ACCESS_RIGHTS, CREATE_CHAN 32 bytes/PV, 32xN=384(if N=12)
				//READ_NOTIFY, EVENT_ADD 48 byte/PV, 32xN=576(if N=12)
				nReqRemainByte = 0;
				while (dataavail(fdSet[i].fdConn))
				{	//480 is a multiple of 32 and 48
					nRequestLen = read( fdSet[i].fdConn, (char *)&bRequestBuf[nReqRemainByte], REQ_BUF_SIZE - nReqRemainByte);	//REQ_BUF_SIZE );
					if ( nRequestLen > 0 )
					{
						nRequestLen = nRequestLen + nReqRemainByte;
						nReqRemainByte = GetRequestRemainByte(bRequestBuf, nRequestLen);
						if (nReqRemainByte < 0)	//illegal value
							break;
						nResponseLen = GetResponse(PROTO_TCP, bRequestBuf, bResponseBuf, nRequestLen - nReqRemainByte);
						if (nResponseLen > 0)
						{	//send response
							if (write( fdSet[i].fdConn, (char *)bResponseBuf, nResponseLen) < 0)
							{
								FD_SET( fdSet[i].fdConn, &error_fds );
								#if DEBUG_CA == 1
								iprintf( "WriteError on fd[%d] GetResponse()\r\n", fdSet[i].fdConn );
								#endif
							}
						}
						nResponseLen = GetReadNotifyResponse(i, bRequestBuf, bResponseBuf, nRequestLen - nReqRemainByte);
						if (nResponseLen > 0)
						{	//send response
							if (write( fdSet[i].fdConn, (char *)bResponseBuf, nResponseLen) < 0)
							{
								FD_SET( fdSet[i].fdConn, &error_fds );
								#if DEBUG_CA == 1
								iprintf( "WriteError on fd[%d] GetReadNotifyResponse()\r\n", fdSet[i].fdConn );
								#endif
							}
						}
						nResponseLen = SetAction(i, bRequestBuf, bResponseBuf, nRequestLen - nReqRemainByte);
						if (nResponseLen > 0)
						{	//send response
							if (write( fdSet[i].fdConn, (char *)bResponseBuf, nResponseLen) < 0)
							{
								FD_SET( fdSet[i].fdConn, &error_fds );
								#if DEBUG_CA == 1
								iprintf( "WriteError on fd[%d] SetAction()\r\n", fdSet[i].fdConn );
								#endif
							}
						}
						if (nReqRemainByte > 0)
							memmove(bRequestBuf, &bRequestBuf[nRequestLen - nReqRemainByte], nReqRemainByte);
					}
					else
					{
						FD_SET( fdSet[i].fdConn, &error_fds );
						#if DEBUG_CA == 1
						iprintf( "ReadError on fd[%d]\r\n", fdSet[i].fdConn );
						#endif
						break; //break while loop
					} //if ( nRequestLen > 0 ) else
				} //while (dataavail(fdSet[i].fdConn))
			} //if ( FD_ISSET( fdSet[i].fdConn, &read_fds ) )
				
			// Check for errors
			if ( FD_ISSET( fdSet[i].fdConn, &error_fds ) )
			{
				close( fdSet[i].fdConn );
				fdSet[i].fdConn = 0;
				PVS_ClearFD(i);	//release PVS with pvSet[i].nFdNdx == nFdNdx
				#if DEBUG_CA == 1
				iprintf( "Error on fd[%d], closing connection\r\n", i );
				#endif
			}
		} //for i loop
   }
}

BOOL IsCheckTimeReached(int nType)
{
	float fNow;
	static long lPreV[2] = { 0, 0 };
	static float fPreV[2] = { 0, 0 };
	fNow = (float)TimeTick / TICKS_PER_SECOND;
	if ((nType == CHECK_DOUBLE) && (fabs(fNow - fPreV[0]) > 0.2))
	{
		fPreV[0] = fNow;
		return TRUE;
	}
	if ((nType == CHECK_STRING) && (fabs(fNow - fPreV[1]) > 0.1))
	{
		fPreV[1] = fNow;
		return TRUE;
	}
	if ((nType == CHECK_ARCHIVE) && (fabs(fNow - (float)lPreV[0]) >= 2))
	{
		lPreV[0] = (long)fNow;
		return TRUE;
	}
	if ((nType == CHECK_LOG) && (fabs(fNow - (float)lPreV[1]) >= 1))
	{	// record must be logged every second
		lPreV[1] = (long)fNow;
		return TRUE;
	}
	return FALSE;
}

void PVCheckTask(void *pd)
{
	//int port = ( int ) pd;
	int nSize, nTotalSize;	// nFdNdx;
	CA_PACKET caResponsePkt;
	int nPvNdx, nPvSetNdx;
	int fdConn, fdConnPrev;
	char szBuffer[960];
	//
	#if DEBUG_CA == 1
	iprintf( "PVCheckTask running\r\n" );
	#endif
	/* The following call to f_enterFS() must be called in every task that accesses
	  the file system.  This must only be called once in each task and must be done before 
	  any other file system functions are used.  Up to 10 tasks can be assigned to use
	  the file system. Any task may also be removed from accessing the file system with a 
	  call to the function f_releaseFS(). */
	//
	/*if f_chdrive() is called in a different task to a
		different drive, it will not affect this task - this task will 
		remain on the same drive.
	*/
	if (f_chdrive( EXT_FLASH_DRV_NUM ) != F_NO_ERROR) //defined in cardtype.h
		iprintf("PVCheck task f_chdrive error\r\n");

	f_chdir( "\\" );
	//
	if (ReadConfigurationFile() == CONFIG_CHANGED)	//read cassetup.ini
	{
		iprintf("Configuration changed, need reboot system...\r\n");
		ShowTitle();
		ShowUser();
		//ForceReboot();
	}
	// read ACF and dabase file here
	//
	while (1)
	{
		FileProcess();	//
		OSTimeDly(2);	//2 = 100 msec, check interval
		IOC_GetAllNewData(); //get ai, di, stringin data into .xNewValue
		if (IsCheckTimeReached(CHECK_ARCHIVE) == TRUE)	{	//send new data to archive engine every 2 seconds (TBD)
			fdConnPrev = -1;
			nTotalSize = 0;
			for (nPvSetNdx = 0; nPvSetNdx < NPVS; nPvSetNdx++)	{
				if (pvSet[nPvSetNdx].nType != PV_TYPE_DOUBLE) //process DOUBLE only
					continue;
				fdConn = fdSet[pvSet[nPvSetNdx].nFdNdx].fdConn;
				nPvNdx = pvSet[nPvSetNdx].nPvNdx;
				if ( fdConn ) { // if (bChanged == TRUE)
					if ((PVS_GetMonitorMask(nPvSetNdx) & DBE_LOG) != 0) { //parameter monitored
						//PV_UpdateOneData(pvSet[nPvSetNdx].nType, nPvNdx); //xNewValue --> xValue, keep bChanged flag unchanged
						nSize = MakePacket_EventAdd(nPvSetNdx, DONT_UPDATE, &caResponsePkt);
						if (((fdConnPrev == -1) || (fdConn == fdConnPrev)) && (nTotalSize + nSize < 960)) 
						{	
							memmove(&szBuffer[nTotalSize], (char *)&caResponsePkt, nSize);
							nTotalSize += nSize;
							fdConnPrev = fdConn;
						}
						else
						{
							write(fdConnPrev, szBuffer, nTotalSize);
							fdConnPrev = fdConn;
							nTotalSize = 0;
							memmove(&szBuffer[nTotalSize], (char *)&caResponsePkt, nSize);
							nTotalSize += nSize;
						}
					} //if ((PVS_GetMonitorMask(nPvSetNdx) & DBE_LOG) != 0)
				} //if ( fdConn )
			}
			if (nTotalSize > 0)
				write(fdConnPrev, szBuffer, nTotalSize);
			//PV_UpdateData(PV_TYPE_DOUBLE); //xNewValue --> xValue, clear bChanged flag
			//PV_UpdateData(PV_TYPE_DOUBLE); //xNewValue --> xValue, clear bChanged flag
		} //if (IsCheckTimeReached(CHECK_DOUBLE) == TRUE)
		//index for PV_DOUBLE or PV_STRING
		if (IsCheckTimeReached(CHECK_DOUBLE) == TRUE) {
			for (nPvSetNdx = 0; nPvSetNdx < NPVS; nPvSetNdx++) {
				if (pvSet[nPvSetNdx].nType != PV_TYPE_DOUBLE)
					continue;
				fdConn = fdSet[pvSet[nPvSetNdx].nFdNdx].fdConn;
				nPvNdx = pvSet[nPvSetNdx].nPvNdx;
				if ( fdConn && PV_DataChanged(PV_TYPE_DOUBLE, nPvNdx)) { // if (bChanged == TRUE)
					if ((PVS_GetMonitorMask(nPvSetNdx) & DBE_VALUE) != 0) { //parameter monitored
						PV_UpdateOneData(PV_TYPE_DOUBLE, nPvNdx); //xNewValue --> xValue, keep bChanged flag unchanged
						nSize = MakePacket_EventAdd(nPvSetNdx, DONT_UPDATE, &caResponsePkt);
						write(fdConn, (char *)&caResponsePkt, nSize);
					}
				}
			}
			PV_UpdateData(PV_TYPE_DOUBLE); //xNewValue --> xValue, clear bChanged flag
		} //if (IsCheckTimeReached(CHECK_DOUBLE) == TRUE)
		if ( IsCheckTimeReached(CHECK_STRING) == TRUE ) 	{
			for (nPvSetNdx = 0; nPvSetNdx < NPVS; nPvSetNdx++) {
				if (pvSet[nPvSetNdx].nType != PV_TYPE_STRING)
					continue;
				fdConn = fdSet[pvSet[nPvSetNdx].nFdNdx].fdConn;
				nPvNdx = pvSet[nPvSetNdx].nPvNdx;
				if (fdConn && PV_DataChanged(PV_TYPE_STRING, nPvNdx)) { // if (NewValue != Value)
					if ((PVS_GetMonitorMask(nPvSetNdx) & DBE_VALUE) != 0) { //parameter monitored
						PV_UpdateOneData(PV_TYPE_STRING, nPvNdx);		//xNewValue --> xValue
						nSize = MakePacket_EventAdd(nPvSetNdx, DONT_UPDATE, &caResponsePkt);
						write(fdConn, (char *)&caResponsePkt, nSize);
					}
				}				
			}	//
			PV_UpdateData(PV_TYPE_STRING); //xNewValue --> xValue, clear bChanged flag
		} //if ( IsCheckTimeReached(CHECK_STRING) == TRUE )
		//set xValue <-- xNewValue, don't change xNewValue
		if (IsCheckTimeReached(CHECK_LOG))
		{
			UpdateLogData();	//write log data into SD card
			SerialOutput();		//serial ouput
		}
	}
	//UnmountExtFlash();  
	f_releaseFS(); 
}

