/*
	File log system
*/
/* Rev:$Revision: 1.3 $ */
/* Copyright $Date: 2007/09/12 07:37:21 $ */ 
 
#include "predef.h" 
#include <stdio.h>
#include <stdlib.h>	//atoi()
#include <ctype.h>
#include <string.h>
#include <basictypes.h>
#include <startnet.h>
#include <time.h>
#include <ucos.h>
#include "ca_proto.h"
#include "cardtype.h"
#include "ffs.h"
//
#define LOG_TIME_PER_WRITE	10
#define YYYYMM_SIZE		6
#define YYYYMMDD_SIZE	8
//
#include <effs_fat/fat.h>
#ifdef USE_MMC
#include <effs_fat/mmc_mcf.h>
#endif
#ifdef USE_CFC
#include <effs_fat/cfc_mcf.h>
#endif
//
#define BUFLEN (F_MAXPATH + F_MAXNAME)
//
extern BOOL bInit;
extern NV_SETTINGS NV_Settings;
//
FILE_PROC_PARAM	FileProcParam;
//
extern "C"
{
   //int get_cd(void);          // Card detection check for MMC cards
   //int get_wp(void);          // Write protection check for MMC cards
}

char szDriveType[20];
extern NV_SETTINGS NV_Settings;
//
char EffsErrorCode[][30] =
{
	"F_NO_ERROR",                // 0
	"F_ERR_INVALIDDRIVE",        // 1   
	"F_ERR_NOTFORMATTED",        // 2
	"F_ERR_INVALIDDIR",          // 3
	"F_ERR_INVALIDNAME",         // 4
	"F_ERR_NOTFOUND",            // 5
	"F_ERR_DUPLICATED",          // 6
	"F_ERR_NOMOREENTRY",         // 7
	"F_ERR_NOTOPEN",             // 8
	"F_ERR_EOF",                 // 9
	"F_ERR_RESERVED",            // 10
	"F_ERR_NOTUSEABLE",          // 11
	"F_ERR_LOCKED",              // 12
	"F_ERR_ACCESSDENIED",        // 13
	"F_ERR_NOTEMPTY",            // 14
	"F_ERR_INITFUNC",            // 15
	"F_ERR_CARDREMOVED",         // 16
	"F_ERR_NODRIVE",             // 17
	"F_ERR_INVALIDSECTOR",       // 18
	"F_ERR_READ",                // 19
	"F_ERR_WRITE",               // 20
	"F_ERR_INVALIDMEDIA",        // 21
	"F_ERR_BUSY",                // 22
	"F_ERR_WRITEPROTECT",        // 23
	"F_ERR_INVFATTYPE",          // 24
	"F_ERR_MEDIATOOSMALL",       // 25
	"F_ERR_MEDIATOOLARGE",       // 26
	"F_ERR_NOTSUPPSECTORSIZE",   // 27
	"F_ERR_DELFUNC",             // 28
	"F_ERR_MOUNTED",             // 29
	"F_ERR_TOOLONGNAME",        // 30 
	"F_ERR_NOTFORREAD",         // 31 
	"F_ERR_DELFUNC",            // 32 
	"F_ERR_ALLOCATION",         // 33 
	"F_ERR_INVALIDPOS",         // 34 
	"F_ERR_NOMORETASK",         // 35 
	"F_ERR_NOTAVAILABLE",       // 36 
	"F_ERR_TASKNOTFOUND",       // 37 
	"F_ERR_UNUSABLE"            // 38 
};

int FindTheOldestDir(char *szDir);
int DeleteDir(char *szDir);
/*-------------------------------------------------------------------
  DisplayEffsErrorCode()
 -------------------------------------------------------------------*/
void DisplayEffsErrorCode(int nCode )
{
	if (nCode <= MAX_EFFS_ERRORCODE )
	{
		iprintf("%s\r\n", EffsErrorCode[nCode]);
	}
	else
	{
		iprintf("Unknown EFFS error code [%d]\r\n", nCode);
	}
}

/*-------------------------------------------------------------------
  DisplayEffsSpaceStats() - Display space used, total and bad
 -------------------------------------------------------------------*/
int DisplayEffsSpaceStats(void)
{
	F_SPACE space;
	volatile int nRet;  
	//iprintf("Retrieving external flash usage...\r\n");
	nRet = f_getfreespace(EXT_FLASH_DRV_NUM, &space);
	
	if (nRet == F_NO_ERROR )
	{
		iprintf("Flash card memory usage (bytes):\r\n");
		iprintf("%s total, ", Make3PNumber(space.total));
		printf("%s free(%.1f%c), ", Make3PNumber(space.free), (float)space.free * 100 / space.total, '%');
		printf("%s used(%.1f%c), ", Make3PNumber(space.used), (float)space.used * 100 / space.total, '%');
		printf("%s bad(%.1f%c)\r\n", Make3PNumber(space.bad), (float)space.bad * 100 / space.total, '%');
	}
	else
	{
		iprintf("\r\n*** Error in f_getfreepace(): ");
		DisplayEffsErrorCode(nRet);
		return F_FAILURE;
	}
	return F_NO_ERROR;
}

int IsCardReady(void)
{
	if (get_cd() == 0 )
	{
		#if DEBUG_FILE == 1 
		iprintf("card is not inserted\r\n");
		#endif
		return F_ERR_CARDREMOVED;
	}
	if (get_wp() == 1 )
	{
		#if DEBUG_FILE == 1
		iprintf("card is write-protected\r\n");
		#endif
		return F_ERR_WRITEPROTECT;
	}
	return F_NO_ERROR;
}

/*-------------------------------------------------------------------
  Init() - Initialization function for setting up EFFS on 
  MultiMedia/Secure Digital cards or Comact Flash cards.
  return SUCCESS or FAILUE
 -------------------------------------------------------------------*/
 
int InitExtFlash(void)
{
	int nRet;
	//
	memset(&FileProcParam, 0, sizeof(FILE_PROC_PARAM)); //parameters for FileProcess();
#if (defined USE_MMC)
	// Card detection check
	if (IsCardReady() != F_NO_ERROR)
		return F_FAILURE;
#endif
   
	/* The f_mountfat() function is called to mount a FAT drive. Although
	there are paramters in this function, they should not be modified. 
	The function call should be used as it appears for a Compact Flash
	card. For reference, the parameters are:
	
	drive_num:   Set to MMC_DRV_NUM, the drive to be mounted
	p_init_func: Points to the initialization function for the drive. 
	           For the Compact Flash drive, the function is located
	           in \nburn\<platform>\system\cfc_mem_nburn.c.
	p_user_info: Used to pass optional information. In this case, the
	           drive number.
	
	The return values are:
		F_NO_ERROR:  drive successfully mounted
		Any other value: Error occurred. See page 22 in the HCC-Embedded
		file system manual for the list of error codes.
	*/ 
	siprintf(szDriveType, "No Drive");   
#if (defined USE_MMC)
	siprintf(szDriveType, "SD/MMC");   
	nRet = f_mountfat(MMC_DRV_NUM, mmc_initfunc, F_MMC_DRIVE0);
#elif (defined USE_CFC)
	siprintf(szDriveType, "CFC");   
	nRet = f_mountfat(CFC_DRV_NUM, cfc_initfunc, F_CFC_DRIVE0);
#endif
	iprintf("Initialize file system .......\r\n");
	if (nRet == F_NO_ERROR )
	{
		iprintf("FAT mount to %s successful\r\n", szDriveType);
	}
	else
	{
		iprintf("FAT mount to %s failed: ", szDriveType);
		DisplayEffsErrorCode(nRet);
		return F_FAILURE;
	}

	/* Change to SD/MMC drive
		We need to call the change function to access the new drive. Note
		that ANY function other than the f_mountfat() is specific to a task.
		This means that if f_chdrive() is called in a different task to a
		different drive, it will not affect this task - this task will 
		remain on the same drive.
	*/
	
	nRet = f_chdrive(EXT_FLASH_DRV_NUM);
	
	if (nRet == F_NO_ERROR )
	{
		iprintf("%s drive change successful\r\n", szDriveType);
	}
	else
	{
		iprintf("%s drive change failed: ", szDriveType);
		DisplayEffsErrorCode(nRet);
		return F_FAILURE;
	}
	DisplayEffsSpaceStats();
	return F_NO_ERROR;
}

/*-------------------------------------------------------------------
BYTE UnmountSD()
 -------------------------------------------------------------------*/
int UnmountExtFlash(void)
{
	int nRet;
	iprintf("Unmounting %s card\r\n\r\n", szDriveType);
	nRet = f_delvolume(EXT_FLASH_DRV_NUM);
	if (nRet != F_NO_ERROR )
	{
		iprintf("*** Error in f_delvolume(): ");
		DisplayEffsErrorCode(nRet);
		return F_FAILURE;
	}
	return F_NO_ERROR;
}

/* 
	create the directory if the directory is not existent
	date time format: YYYYMMDD_hhmmss
	return F_NO_ERROR or F_FAILURE
*/
int GoToDirectory(char *szDateTime)
{
	char szNewDir[30];
	char szCurDir[30];
	//
	strncpy(szNewDir, szDateTime, YYYYMM_SIZE);	//YYYYMM
	szNewDir[YYYYMM_SIZE] = 0;
	if (f_getcwd(szCurDir, 29) != F_NO_ERROR)
		return F_FAILURE;
	if (strcmp(szCurDir, szNewDir) == 0)	//if directory is the same, return
		return F_NO_ERROR;

	if (f_chdir("\\") != F_NO_ERROR)	// go to root first
		return F_FAILURE;
	if (f_chdir(szNewDir) != F_NO_ERROR)
	{
		if (f_mkdir(szNewDir) != F_NO_ERROR)
			return F_FAILURE;
		f_chdir(szNewDir);
	} 
	return F_NO_ERROR;
}
/*
	szNewFile, szNewDate, szOldDate format: YYYYMMDD_hhmmss
	change to new log file when day changes
*/
int IsNewMonth(char *szNewFile)
{
	char szNewDate[30];
	static char szOldDate[30] = {0};
	
	//GetSystemTime(szNewDate, DATE_TIME_COMPACT);	//format: YYYYMMDD_hhmmss
	GetUTCTime(szNewDate, DATE_TIME_COMPACT);	//format: YYYYMMDD_hhmmss
	if (strncmp(szNewDate, szOldDate, YYYYMM_SIZE) == 0) //YYYYMM -> different month
	{
		strcpy(szNewFile, szNewDate);	//YYYYMMDD
		szNewFile[YYYYMMDD_SIZE] = 0;
		return F_IS_SAME_MONTH;	//no new file needed
	}
	strcpy(szNewFile, szNewDate);	//YYYYMMDD
	szNewFile[YYYYMMDD_SIZE] = 0;
	strcpy(szOldDate, szNewFile);
	return F_IS_NEW_MONTH;	//new file needed
}

int WriteLogData(BYTE *pData, int nBytes)
{
	char szNewFile[30];		//"20090918_010203"
	char szFileName[30];	//filename = "20090918_010203.bin"
	char szNameList[128];
	int nRet, nHeaderSize;
	int nNewFile;
	F_FIND finder;    // location to store the information retreived
	F_SPACE space;
	unsigned long ulBytesPerMonth;
	//
	if (IsCardReady() != F_NO_ERROR)
		return F_FAILURE;
	if (NV_Settings.nLogFormat == FORMAT_NO_LOG)
		return F_FAILURE;
	if (IsNewMonth(szNewFile) == F_IS_NEW_MONTH) { //change to new month
		if (NV_Settings.nAutoDelete == 1) { //new directory needed, check capacity
			ulBytesPerMonth = (unsigned long)nBytes * 86400 / NV_Settings.nLogInterval;
			ulBytesPerMonth = ulBytesPerMonth / LOG_TIME_PER_WRITE * 35;
			while (1)
			{
				nRet = f_getfreespace(EXT_FLASH_DRV_NUM, &space);
				if (nRet != F_NO_ERROR)
					break;	
				if (space.free < ulBytesPerMonth) { //48*86400*30=124,416,000
					#if DEBUG_FILE == 1
					printf("disk full, free = %ld, bytes per month = %ld\r\n", space.free, ulBytesPerMonth);
					#endif
					if (FindTheOldestDir(szFileName) == F_NO_ERROR)
					{
						if (DeleteDir(szFileName) != F_NO_ERROR) 
							break;	//delete the oldest directory
					}
				}
				else
					break;
			}
			nRet = f_getfreespace(EXT_FLASH_DRV_NUM, &space);
			#if DEBUG_FILE == 1
			printf("=================================\r\n");
			printf("free = %s, ", Make3PNumber(space.free));
			printf("bytes per month = %s\r\n", Make3PNumber(ulBytesPerMonth));
			#endif
		} //if (NV_Settings.nAutoDelete == 1)
	} //if (IsNewMonth(szNewFile) == F_IS_NEW_MONTH)
	
	if (NV_Settings.nLogFormat == FORMAT_TEXT)
		sprintf(szFileName, "%s.txt", szNewFile);
	else	//FORMAT_BINARY
		sprintf(szFileName, "%s.bin", szNewFile);
	//
	if (GoToDirectory(szNewFile) == F_FAILURE)
	{
		#if DEBUG_FILE == 1
		iprintf("GoToDirectory() fail\r\n");
		#endif
		return F_FAILURE;
	}
	// 

	/* Find first file or subdirectory in specified directory. First call the
		f_findfirst function, and if file was found get the next file with 
		f_findnext function. Files with the system attribute set will be ignored.
		Note: If f_findfirst() is called with "*.*" and the current directory is 
		not the root directory, the first entry found will be "." - the current 
		directory.
	*/
	volatile int rc = f_findfirst(szFileName, &finder);
	if (rc == F_NO_ERROR )  // found a file or directory
		nNewFile = F_IS_SAME_FILE;
	else
		nNewFile = F_IS_NEW_FILE;
	
	F_FILE* fp = f_open(szFileName, "a+"); //append file, long file name
	if (fp )
	{
		if (nNewFile == F_IS_NEW_FILE)	//write header
		{
			nHeaderSize = PV_GetNameList(PV_TYPE_DOUBLE, szNameList, NV_Settings.nLogFormat);
			nRet = f_write(szNameList, 1, nHeaderSize, fp); 
		}
		nRet = f_write(pData, 1, nBytes, fp); 
		if (nRet != nBytes )
		{
			#if DEBUG_FILE == 1
			iprintf("\r\n*** Error in f_write(%s): %d out of %d bytes written\r\n", szFileName, nRet, nBytes);
			#endif
		}
		
		nRet = f_close(fp);  // Close a previously opened file of type F_FILE
		if (nRet != F_NO_ERROR )
		{
			#if DEBUG_FILE == 1
			iprintf("*** Error in f_close(%s): ", szFileName);
			#endif
			DisplayEffsErrorCode(nRet);
			return F_FAILURE;
		}
	}
	else
	{
		#if DEBUG_FILE == 1
		iprintf("Failed to open file %s\r\n", szFileName);
		#endif
		return F_FAILURE;
	}
	return F_NO_ERROR;
}

int UpdateLogData()
{
	static BYTE pLogData[1024];
	static int nLogNdx = 0, nLogDataSize;
	static int nTime = 0;
	static time_t t_prev = 0;
	time_t t_now;
	//
	if (bInit == FALSE)
		return 0;
	t_now = time(NULL);
	if (t_now - t_prev < NV_Settings.nLogInterval)
		return 0;
	t_prev = t_now;
	//
	nLogDataSize = PV_GetDataList(PV_TYPE_DOUBLE, (char *)&pLogData[nLogNdx], NV_Settings.nLogFormat);
	if (nLogDataSize == 0)	// no data is logged
		return 0;
	nLogNdx = nLogNdx + nLogDataSize;
	nTime += NV_Settings.nLogInterval;
	if ((nTime >= LOG_TIME_PER_WRITE) || (nLogNdx > 896))	//write log every 10 seconds
	{
		//#if DEBUG_FILE == 1
		//iprintf("write log data %ld\r\n", (long)t_prev); //second since 1970/1/1
		//#endif
		//---------- critical process start ---------------
		OS_CRIT cs; 	         // The critical section identifer
   		OSCritInit(&cs);
   		OSCritEnter(&cs, 0);
		WriteLogData(pLogData, nLogNdx);
		OSCritLeave(&cs);
		//---------- critical process end ---------------
		nLogNdx = 0;
		nTime = 0;
		return 2;
	}
	return 1;
}

void SerialOutput(void)
{
	char szBuffer[256];
	if (NV_Settings.nSerialOutput == 0)
		return;
	PV_GetDataList(PV_TYPE_DOUBLE, szBuffer, FORMAT_TEXT);
	printf(szBuffer);
}
/*
	Search the oldest directory with format YYYYMM
	return F_NO_ERROR if directory is found
*/
int FindTheOldestDir(char *szDir)
{
	F_FIND finder;    // location to store the information retreived
	char szOldestDir[10] = "999999";
	BOOL bFound = FALSE;
	//
	volatile int rc = f_findfirst("*.*", &finder);
	if (rc == F_NO_ERROR ) { // found a file or directory
		do {
			if ((finder.attr & F_ATTR_DIR ) != 0) { 
				if ((strcmp(szOldestDir, finder.filename ) > 0) && (strlen(finder.filename) == YYYYMM_SIZE)) {
					strcpy(	szOldestDir, finder.filename);
					bFound = TRUE;
				}
	        }
		}
		while (!f_findnext(&finder ));
	}
	if (bFound == TRUE)	{
		strcpy(szDir, szOldestDir);
		return F_NO_ERROR;
	}
	else
		return F_FAILURE;
}

int DeleteFiles()
{
	F_FIND finder;    // location to store the information retreived
	//
	volatile int rc = f_findfirst("*.*", &finder);
	if (rc == F_NO_ERROR ) { // found a file or directory
		do {
			if ((finder.attr & F_ATTR_ARC ) != 0) { //???
				f_delete(finder.filename);	
	        }
		}
		while (!f_findnext(&finder ));
	}
	return rc;
}

// delete directory in root
int DeleteDir(char *szDir)
{
	int nRet;
	if (f_chdir(szDir) != F_NO_ERROR)
		return F_FAILURE;
	DeleteFiles();
	f_chdir("\\");		
	nRet = f_rmdir(szDir);
	if (nRet == F_NO_ERROR)
		printf("Directory %s is successfully removed.\r\n", szDir);
	else
		printf("Directory deletion error\r\n");
	return nRet;
} 
/*-------------------------------------------------------------------
  DumpDir() - Displays a list of all directories and files in the 
  file system.
  !!! this is a recursive function
 -------------------------------------------------------------------*/
int DumpDir(void)
{
	F_FIND finder;    // location to store the information retreived
	char szBuffer[BUFLEN]; 
	
	/* Find first file or subdirectory in specified directory. First call the
		f_findfirst function, and if file was found get the next file with 
		f_findnext function. Files with the system attribute set will be ignored.
		Note: If f_findfirst() is called with "*.*" and the current directory is 
		not the root directory, the first entry found will be "." - the current 
		directory.
	*/
	//
	volatile int rc = f_findfirst("*.*", &finder);
	if (rc == F_NO_ERROR )  // found a file or directory
	{
		do {
			if ((finder.attr & F_ATTR_DIR ) )	{	
				if (finder.filename[0] != '.' ) {
					//iprintf("Directory [%s]\r\n", finder.filename); //finder.filename);
					f_chdir(finder.filename);
					f_getcwd(szBuffer, BUFLEN);
					iprintf("Directory [%s]\r\n", szBuffer); //finder.filename);
					DumpDir();	//recursive function
					f_chdir(".."); //go back to upper directory
				}
	        }
	        else {
	        	f_getcwd(szBuffer, BUFLEN);
	        	iprintf("%s\\%13s : %10s bytes\r\n", szBuffer, finder.filename, Make3PNumber(finder.filesize));
			}
		}
		while (!f_findnext(&finder ));
	}
	//
	return rc;
}

int DumpOneDir(char *szDir)
{
	F_FIND finder;    // location to store the information retreived
	char szBuffer[BUFLEN]; 
	
	/* Find first file or subdirectory in specified directory. First call the
		f_findfirst function, and if file was found get the next file with 
		f_findnext function. Files with the system attribute set will be ignored.
		Note: If f_findfirst() is called with "*.*" and the current directory is 
		not the root directory, the first entry found will be "." - the current 
		directory.
	*/
	//
	if (f_chdir(szDir) != F_NO_ERROR)
		return F_FAILURE;
	volatile int rc = f_findfirst("*.*", &finder);
	if (rc == F_NO_ERROR )  // found a file or directory
	{
		do {
			f_getcwd(szBuffer, BUFLEN);
			if ((finder.attr & F_ATTR_DIR))	{
				if (finder.filename[0] != '.')	
					iprintf("Directory [%s] %s\r\n", szBuffer, finder.filename); //finder.filename);
	        }
	        else {
	        	iprintf("%s\\%13s : %10s bytes\r\n", szBuffer, finder.filename, Make3PNumber(finder.filesize));
			}
		}
		while (!f_findnext(&finder ));
	}
	//
	return rc;
}

int DumpOnlyDir(void)
{
	F_FIND finder;    // location to store the information retreived
	//
	iprintf("Directory\r\n");
	volatile int rc = f_findfirst("*.*", &finder);
	if (rc == F_NO_ERROR )  // found a file or directory
	{
		do {
			if ((finder.attr & F_ATTR_DIR))	{
				if (finder.filename[0] != '.')	
					iprintf("%s\r\n", finder.filename); //finder.filename);
	        }
		}
		while (!f_findnext(&finder ));
	}
	//
	return rc;
}
/*
 * szBuffer = "NTP=172.16.1.201\n"
 * szName = "NTP="
 * szResult = "172.16.1.201"
*/
int GetConfigName(char *szBuffer, char *szName, char *szResult)
{
	char *p1, *p2;
	char *p;
	//
	p = szBuffer;
	do
	{
		p1 = strstr(p, szName);
		p = p1 + 1;		//find next
		if (p1 == NULL)	//if szName can't be found in szBuffer
			break;
		if (*(p1 - 1) == '/')	//parameter is commented
			continue;	//find next
		p1 = p1 + strlen(szName); //NTP=, move to the char after =
		p2 = strchr(p1, '\n');
		if (p2 == NULL)		//illegal ending
			continue;
		//
		memmove(szResult, p1, p2 - p1);
		szResult[p2 - p1] = 0; //remove '\n'
		return SUCCESS;
	} while (p1 != NULL);
	return FAILURE;
}

int GetConfigInt(char *szBuffer, char *szName, int *npI, int nMax, int nMin, int nDef)
{
	char szResult[20];
	int nTemp, nRet;
	nRet = CONFIG_FAILURE;
	if (GetConfigName(szBuffer, szName, szResult) == SUCCESS)
	{
		nTemp = atoi(szResult);
		if (nTemp != *npI)
		{
			*npI = nTemp;
			nRet = CONFIG_CHANGED;
		}
		else
			nRet = CONFIG_NO_CHANGE;
	}
	if (*npI > nMax || *npI < nMin)
	{
		*npI = nDef;
		nRet = CONFIG_CHANGED;
	}
	//value not changed if it is not in the list
	return nRet;
}

int GetConfigIP(char *szBuffer, char *szName, IPADDR *piaV)
{
	char szResult[32];
	IPADDR iaTemp;
	if (GetConfigName(szBuffer, szName, szResult) == SUCCESS)
	{
		iaTemp = AsciiToIp(szResult);
		if (iaTemp != *piaV)
		{
			*piaV = iaTemp;
			return CONFIG_CHANGED;
		}
		return CONFIG_NO_CHANGE;
	}
	if (*piaV == 0)
		return CONFIG_NO_CHANGE;
	*piaV = 0; //changed to 0.0.0.0 if it is not found in the list
	return CONFIG_CHANGED;
}
/*
 * szBuffer = "NTP=172.16.1.201\n"
 * szName = "NTP="
 * szResult = "172.16.1.201"
*/
int GetConfigStr(char *szBuffer, char *szName, char *szStr)
{
	char szResult[32];
	if (GetConfigName(szBuffer, szName, szResult) == SUCCESS)
	{
		if (strcmp(szStr, szResult) != 0) { // if a new value in SD card file
			strcpy(szStr, szResult);
			return CONFIG_CHANGED;
		}
		return CONFIG_NO_CHANGE; //szStr is not changed
	}
	if (szStr[0] == 0)
		return CONFIG_NO_CHANGE;
	szStr[0] = 0; //clear szStr if not found
	return CONFIG_CHANGED;
}

int ReadConfigurationFile(void)
{
	char szFileName[13] = "cassetup.ini";
	char szName[20];
	char szBuffer[MAX_CONFIG_LEN];
	int nConfigChanged = 0; 
	int i;
	//
	if (f_chdir("\\") != F_NO_ERROR )
		return CONFIG_FAILURE;
	//
	F_FILE *fp = f_open(szFileName, "r");
	long lSize = f_filelength(szFileName);
	//
	if (fp == NULL) //if "cassetup.ini" is not found
		return CONFIG_FAILURE;
	//
	if (lSize > MAX_CONFIG_LEN - 2)
		lSize = MAX_CONFIG_LEN - 2;
	//
	f_read(szBuffer, lSize, 1, fp);
	strcat(szBuffer, "\n");
	PackChar(szBuffer, "\r ");
	//pack space
	nConfigChanged |= GetConfigStr(szBuffer, "DeviceName=", NV_Settings.szDeviceName);
	nConfigChanged |= GetConfigIP(szBuffer, "NTP=", &NV_Settings.iaNTP); //need reboot
	nConfigChanged |= GetConfigIP(szBuffer, "NetMask=", &NV_Settings.iaNetMask);
	nConfigChanged |= GetConfigInt(szBuffer, "LogFormat=", &NV_Settings.nLogFormat, 
		FORMAT_TEXT, FORMAT_NO_LOG, FORMAT_BINARY);
	nConfigChanged |= GetConfigInt(szBuffer, "LogInterval=", &NV_Settings.nLogInterval,
		1200, 1, 1);
	for (i = 0; i < MAX_AUTHORIZE_NUM; i++)
	{
		sprintf(szName, "AuIP%d=", i + 1); 
		nConfigChanged |= GetConfigIP(szBuffer, szName, &NV_Settings.iaAuthorized[i]);
		sprintf(szName, "AuRW%d=", i + 1); 
		nConfigChanged |= GetConfigInt(szBuffer, szName, (int *)&NV_Settings.nAuthorizedRW[i],
			3, 0, 3);	//default RW(3), R(1), W(2)
		sprintf(szName, "Host%d=", i + 1); //Host1=hist1:user1,user2
		nConfigChanged |= GetConfigStr(szBuffer, szName, NV_Settings.szAuthorizedHC[i]);
	}
	// UTC time = local time - nTZOffset * 60
	nConfigChanged |= GetConfigInt(szBuffer, "TimeZone=", &NV_Settings.nTZOffset, 
		720, -660, 480);	//default Taiwan time, +8 hours
	//nConfigChanged |= GetConfigInt(szBuffer, "SOUT=", &NV_Settings.nSerialOutput, 
	//	1, 0, 0);	//default no serial output after power on process
	NV_Settings.nSerialOutput = 0; //always 0 at power on stage
	nConfigChanged |= GetConfigInt(szBuffer, "AutoDel=", &NV_Settings.nAutoDelete, 
		1, 0, 1);	//default auto delete enabled
	nConfigChanged |= GetConfigInt(szBuffer, "NTPInterval=", &NV_Settings.nNTPInterval, 
		14400, 1, 2);	//default NTP calbration per day
	f_close(fp);
	nConfigChanged &= CONFIG_CHANGED;
	if (nConfigChanged == CONFIG_CHANGED)
	{
		iprintf("Read new configuration from SD card\r\n");
		SaveUserParameters(&NV_Settings, sizeof(NV_SETTINGS ));
	}
	return nConfigChanged;
}
/*
	Read access configuration file cassetup.acf
UAG(uag) {user1,user2}
HAG(hag) {host1,host2}
ASG(DEFAULT) {
	RULE(1,READ)
	RULE(1,WRITE) {
		UAG(uag)
		HAG(hag)
	}
}
Read access to anyone located anywhere.
Write access to user1 and user2 if they are located at host1 or host2. 
*/
void ReadACF(void)
{	//user: CLIENT_NAME, host: HOST_NAME
	//host1:user1(RW),user2(R)
	//host2:user1(RW),user3(R)
}

void FileProcess(void)
{
	int nCode;
	if (FileProcParam.bDeleteDir == TRUE)
	{
		FileProcParam.bDeleteDir = FALSE;
		nCode = f_chdir("\\");
		if ( nCode != F_NO_ERROR) {
			DisplayEffsErrorCode(nCode);
			return;
		}
		DeleteDir(FileProcParam.szDir);
		DisplayEffsSpaceStats();
	}
	if (FileProcParam.bDumpDir == TRUE)
	{
		FileProcParam.bDumpDir = FALSE;
		if (FileProcParam.szDir[0] == 0) {
			nCode = f_chdir("\\");
			DumpDir();
		}
		else if (FileProcParam.szDir[0] == '*') {
			nCode = f_chdir("\\");
			DumpOnlyDir();
		}
		else {
			nCode = f_chdir("\\");
			DumpOneDir(FileProcParam.szDir);
		}
		if (nCode != F_NO_ERROR) {
			DisplayEffsErrorCode(nCode);
			return;
		}
		DisplayEffsSpaceStats();
	}
	if (FileProcParam.bFormat == TRUE)
	{
		FileProcParam.bFormat = FALSE;
		nCode = f_format(MMC_DRV_NUM, F_FAT32_MEDIA);
		if (nCode == F_NO_ERROR)
			iprintf("drive FAT32 successfully formatted\r\n");
		else
			DisplayEffsErrorCode(nCode);
		DisplayEffsSpaceStats();		
	}
}
