/*
	User parameters
*/
#include "predef.h"
#include <stdio.h>
#include <stdlib.h>
#include <startnet.h>		//TICKS_PER_SECOND
#include <basictypes.h>
#include <string.h>
#include <ctype.h>
#include <system.h> // Found in C:\Nburn\include, SaveParameters
#include <utils.h> // Found in C:\Nburn\include
#include <nbtime.h>
#include <rtc.h>
#include <bsp.h>	//roboot system
#include <math.h>	//fabs()
#include <ucos.h>
#include <udp.h>
#include "ca_proto.h"
#include "ffs.h"

// This key can be updated if you modify your structure. For example,
// you may need to do something special if you load a new version 
// of s/w in a device that has the old structure values. 
#define VERIFY_KEY (0x48666050)  // NV Settings key code

extern CAFD_SET fdSet[NFDS];	//
extern PV_SET pvSet[NPVS];

PV_DOUBLE pvDouble[PV_DBL_NUM];
PV_STRING pvString[PV_STR_NUM];
PV_SHORT pvShort[PV_SHORT_NUM];
//
NV_SETTINGS NV_Settings;
BOOL bCheckedNV;
BOOL bSendBeacon = 1;
int nTxInterface = FORMAT_TEXT;
extern UDP_CLIENT UdpClient;

extern FILE_PROC_PARAM FileProcParam;	//ffs.h

void CheckNVSettings(void)
{
	BYTE bSW;
	char szBuffer[20]; 
	iprintf("Checking NV_Settings if Flash...\r\n");
	iprintf("Size of NV Structure: %d\r\n", sizeof(NV_Settings));
	
	NV_SETTINGS *pData = (NV_SETTINGS *) GetUserParameters();
	iprintf("Verify key = 0x%lX\r\n", pData->dwVerifyKey);
	
	if (pData->dwVerifyKey != VERIFY_KEY)
	{
		iprintf("Flash verification key is INVALID - Initializing Flash\r\n");
		memset(&NV_Settings, 0, sizeof(NV_SETTINGS));
		NV_Settings.dwVerifyKey = VERIFY_KEY;
		bSW = getdipsw();
		sprintf(szBuffer, "nbcas%03d", (int)bSW);
		strcpy(NV_Settings.szDeviceName, szBuffer);	//Netburner CAS
		// 172.16.1.xxx
		NV_Settings.iaAuthorized[0] = AsciiToIp("172.16.1.255");
		NV_Settings.nAuthorizedRW[0] = CA_PROTO_ACCESS_RIGHT_RW;
		NV_Settings.iaNTP = AsciiToIp("172.16.1.201");
		NV_Settings.iaNetMask = AsciiToIp("255.255.0.0");
		NV_Settings.nLogFormat = FORMAT_BINARY;
		NV_Settings.nLogInterval = 1;
		NV_Settings.nSerialOutput = 0;
		NV_Settings.nNTPInterval = 1440; //NTP refresh every day, unit minute 
		SaveUserParameters(&NV_Settings, sizeof(NV_SETTINGS));
	}
	else
	{
		iprintf("Flash verification is VALID - reading values from Flash\r\n");
		memmove((BYTE *)&NV_Settings, (BYTE *)pData, sizeof(NV_SETTINGS));
	}
	if (NV_Settings.nLogFormat < FORMAT_NO_LOG || NV_Settings.nLogFormat > FORMAT_TEXT)
		NV_Settings.nLogFormat = FORMAT_BINARY;
	if (NV_Settings.nLogInterval < 1 || NV_Settings.nLogInterval > 1200)
		NV_Settings.nLogInterval = 1;
	//NV_Settings.iaNTP = AsciiToIp("172.16.1.204"); //testing
	bCheckedNV = TRUE;
}

#define NTP_PKT_LEN		48
#define NTP_PORT		123
/*typedef struct NTP_PKT
{
	BYTE bFlag;					//0xcc
	BYTE bPeerClkStratum;
	BYTE bPeerPollInterval;
	BYTE bPeerClockPrecision;	//4
	float fRootDelay;			//8
	float fRootDispersion;		//12
	DWORD dwClockID;			//16
	double dReferenceClok;		//24
	double dOriginalTimeStamp;	//32
	double dReceiveTimeStamp;	//40
	double dTransmitTimestamp;	//48
} NTP_PKT; */

void GetDateTimeString(char *szDateTime)
{
	char szBuffer[32];
	GetSystemTime(szBuffer, DATE_TIME_NORMAL);
	sprintf(szDateTime, "system time = %s\r\n", szBuffer);
	GetUTCTime(szBuffer, DATE_TIME_NORMAL);
	sprintf(szDateTime, "%sUTC time = %s\r\n", szDateTime, szBuffer);
}

BOOL MySetNTPTime(IPADDR iaNTP)
{
	BYTE bBuffer[NTP_PKT_LEN] = { 0x0b };
	time_t t;
	int nLen;	
	//
	{
		UDPPacket UdpTxPkt;
		UdpTxPkt.SetSourcePort( NTP_PORT );
		UdpTxPkt.SetDestinationPort( NTP_PORT );
		UdpTxPkt.AddData(bBuffer, NTP_PKT_LEN);
		UdpTxPkt.Send(iaNTP);
	}
	OS_FIFO fifo;
	OSFifoInit( &fifo );

	// Register to listen for UDP packets on port number 'port'
	//!!!don't use RegisterMulticastFifo(), it is not working
	RegisterUDPFifo(NTP_PORT, &fifo);
	// We construct a UDP packet object using the FIFO. 
	// This constructor will only return when we have received a packet
	UDPPacket upkt( &fifo, 10); // /* Replace this 0 with a tick count to have a time out delay

	// Did we get a valid packet or just time out?
	if ( upkt.Validate())
	{
		nLen = (int)upkt.GetDataSize();
		if (nLen != NTP_PKT_LEN)
			return FALSE;
		memmove(bBuffer, upkt.GetDataBuffer(), nLen );
		t = *(time_t *)&bBuffer[NTP_PKT_LEN - 16];			//seconds UTC since 1900/1/1
		t = t - (DWORD)SEC_FOR_70Y + (DWORD)NV_Settings.nNTPInterval * 60;	//seconds UTC since 1970/1/1
		if (t < SEC_2009Y) //|| (t > SEC_2039Y))
			return FALSE;	//invalid time 
		set_time(t); // 
		return TRUE;
	}
	return FALSE;
}
/*
 get from NTP server or RTC 
*/
void TimeCalibration(void)
{
	int i;
	char szDateTime[64];
	struct tm *p;
	time_t t_new = 0, t_prev;
	BOOL bSet;
	//Set the time using NTP
	//available NTP server: 
	//tock.stdtime.gov.tw 
	//time.stdtime.gov.tw		220.130.158.52 
	//clock.stdtime.gov.tw 
	//freq_f.stdtime.gov.tw 
	//tick.stdtime.gov.tw 
	//*
	//RTCGetTime(bts);
	//t_prev = mktime(&bts);
	t_prev = time(NULL); //get local time
	for (i = 0; i < 3; i++)
	{
		// Gets time from an NTP Server
		#if DEBUG_TIME == 1
		iprintf("Try to get NTP time. Next sync. %ld sec later.\r\n", (long)NV_Settings.nNTPInterval * 60);
		#endif
		//bSet = SetNTPTime(NV_Settings.iaNTP); //this is UTC time
		bSet = MySetNTPTime(NV_Settings.iaNTP); //this is UTC time
		if (bSet == FALSE)
		{
			#if DEBUG_TIME == 1
			iprintf("Get NTP failed, waiting 2 seconds to try again\r\n");
			#endif
			OSTimeDly(TICKS_PER_SECOND * 2);
		}
		else
		{
			#if DEBUG_TIME == 1
			iprintf("Got NTP time\r\n");
			#endif
			//time() has change to UTC time due to SetNTPTime() function
			t_new = time(NULL) + (long)NV_Settings.nTZOffset * 60; //change to local time
			set_time(t_new);  //time() is now back to local time
			if (abs(t_new - t_prev) >= 3)
			{
				iprintf("RTC time correction\r\n");
				p = localtime(&t_new);
				RTCSetTime(*p);
			} 
			break;
		}
	}
	//*/
	GetDateTimeString(szDateTime);
	iprintf(szDateTime);
}

void PV_UpdateChange(int nType, int nPvNdx, char *szValue)
{
	switch (nType)
	{
		case PV_TYPE_SHORT:
			pvShort[nPvNdx].sValue = pvShort[nPvNdx].sNewValue;
			break;
		case PV_TYPE_DOUBLE:
			pvDouble[nPvNdx].dValue = pvDouble[nPvNdx].dNewValue;
			break;
		case PV_TYPE_STRING:
			strcpy(pvString[nPvNdx].szValue, pvString[nPvNdx].szNewValue);
			break;
	}
}
/*
 * get xValue, don't update xNewValue
	return 
		data length of payload.
		pPayload: payload data
*/
int PV_GetPayload(WORD wType, int nPvNdx, BYTE *pPayload)
{
	int nPayloadSize = 0;
	struct PL_DBR_TIME_SHORT plTimeShort;	//PL=payload
	struct PL_DBR_TIME_DOUBLE plTimeDouble;	//PL=payload
	struct PL_DBR_TIME_STRING plTimeString;
	time_t time_val = time(NULL);
	//input time_val = second from 1970/1/1 12:00:00, local time
	//output time = second from 1990/1/1 12:00:00, UTC time
	//time_val is local time
	time_val = ChangeToUTC1990(time_val);
	//
	if (wType == DBR_CTRL_DOUBLE)	//(34), necessary for archive engine request
	{
		nPayloadSize = sizeof(PL_DBR_CTRL_DOUBLE);
		pvDouble[nPvNdx].dbrCtrl.dValue = pvDouble[nPvNdx].dValue; //current value
		memmove(pPayload, (BYTE *)&pvDouble[nPvNdx].dbrCtrl, nPayloadSize);
	}
	if (wType == DBR_TIME_SHORT)
	{
		//update dValue or szValue from xNewValue
		plTimeShort.wStatus = 0x0000;
		plTimeShort.wSeverity = 0x0000;
		plTimeShort.dwTime1 = time_val;
		plTimeShort.dwTime2 = 0;	//precise time TBD
		plTimeShort.sRISC_PAD = 0; //0x20430000;	//TBD
		plTimeShort.sValue = pvShort[nPvNdx].sValue;
		nPayloadSize = sizeof(PL_DBR_TIME_SHORT);	//16
		memmove(pPayload, (BYTE *)&plTimeShort, nPayloadSize);
	}
	if (wType == DBR_TIME_DOUBLE)
	{
		//update dValue or szValue from xNewValue
		plTimeDouble.wStatus = 0x0000;
		plTimeDouble.wSeverity = 0x0000;
		plTimeDouble.dwTime1 = time_val;
		plTimeDouble.dwTime2 = 0;	//precise time TBD
		plTimeDouble.dwRISC_PAD = 0; //0x20430000;	//TBD
		plTimeDouble.dValue = pvDouble[nPvNdx].dValue;
		nPayloadSize = sizeof(PL_DBR_TIME_DOUBLE);	//24
		memmove(pPayload, (BYTE *)&plTimeDouble, nPayloadSize);
	}
	if (wType == DBR_TIME_STRING)
	{
		//update dValue or szValue from xNewValue
		//IOC_GetData(nPvNdx, PV_TYPE_STRING, DO_UPDATE);
		plTimeString.wStatus = 0x0000;
		plTimeString.wSeverity = 0x0000;
		plTimeString.dwTime1 = time_val;
		plTimeString.dwTime2 = 0;
		//memset(pPayload, 0, MAX_STRING_SIZE);
		strcpy(plTimeString.szValue, pvString[nPvNdx].szValue);
		//nPayloadSize = Make8(12 + Make8(strlen(plTimeString.szValue)));
		nPayloadSize = Make8(12 + strlen(plTimeString.szValue) + 1);	//plus 1 is important!!!
		memmove(pPayload, (BYTE *)&plTimeString, nPayloadSize); //0-terminated
	}
	return nPayloadSize;
}
/*
 * compare xNewValue and xValue
 * */
BOOL PV_DataChanged(int nType, int nPvNdx)
{
	int nRet = 0;
	//if (pvDouble[nPvNdx].bAccessRight = CA_PROTO_ACCESS_RIGHT_WRITE) //write only
	//	return FALSE;	//can't change
	switch (nType)
	{
		case PV_TYPE_SHORT:
			if (fabs((float)(pvShort[nPvNdx].sNewValue - pvShort[nPvNdx].sValue)) > pvShort[nPvNdx].dHysteresis)
				pvShort[nPvNdx].bChanged = TRUE;	//value changed
			return pvShort[nPvNdx].bChanged;
		case PV_TYPE_DOUBLE:
			if (fabs(pvDouble[nPvNdx].dNewValue - pvDouble[nPvNdx].dValue) > pvDouble[nPvNdx].dHysteresis)
				pvDouble[nPvNdx].bChanged = TRUE;	//value changed
			return pvDouble[nPvNdx].bChanged;
		case PV_TYPE_STRING:
			nRet = strcmp(pvString[nPvNdx].szNewValue, pvString[nPvNdx].szValue);
			if (nRet != 0)
				pvString[nPvNdx].bChanged = TRUE;
			return pvString[nPvNdx].bChanged;
	}
	return FALSE;
}

// set xValue from xNewValue
void PV_UpdateOneData(int nType, int nPvNdx)
{
	switch (nType)
	{
		case PV_TYPE_SHORT:
			pvShort[nPvNdx].sValue = pvShort[nPvNdx].sNewValue;
			break;
		case PV_TYPE_DOUBLE:
			pvDouble[nPvNdx].dValue = pvDouble[nPvNdx].dNewValue;
			break;
		case PV_TYPE_STRING:
			strcpy(pvString[nPvNdx].szValue, pvString[nPvNdx].szNewValue);
			break;
	}
}

// set all xValue from xNewValue
// clear bChanged flag
void PV_UpdateData(int nType)
{
	int nPvNdx;
	switch (nType)
	{
		case PV_TYPE_SHORT:
			for (nPvNdx = 0; nPvNdx < PV_SHORT_NUM; nPvNdx++) {
				pvShort[nPvNdx].sValue = pvShort[nPvNdx].sNewValue;
				pvShort[nPvNdx].bChanged = FALSE;
 			}
			break;
		case PV_TYPE_DOUBLE:
			for (nPvNdx = 0; nPvNdx < PV_DBL_NUM; nPvNdx++) {
				pvDouble[nPvNdx].dValue = pvDouble[nPvNdx].dNewValue;
				pvDouble[nPvNdx].bChanged = FALSE;
			}
			break;
		case PV_TYPE_STRING:
			for (nPvNdx = 0; nPvNdx < PV_STR_NUM; nPvNdx++) {
				strcpy(pvString[nPvNdx].szValue, pvString[nPvNdx].szNewValue);
				pvString[nPvNdx].bChanged = FALSE;
			}
			break;
	}
}

WORD PV_GetAccessRight(int nType, int nPvNdx)
{
	switch (nType)
	{
		case PV_TYPE_SHORT:
			return pvShort[nPvNdx].wAccessRight;
		case PV_TYPE_DOUBLE:
			return pvDouble[nPvNdx].wAccessRight;
		case PV_TYPE_STRING:
			return pvString[nPvNdx].wAccessRight;
	}
	return 0;	// no read/write function
}

WORD PV_GetDataType(int nType)
{
	WORD wDataType;
	wDataType = (nType == PV_TYPE_DOUBLE) ? DBR_DOUBLE :
		(nType == PV_TYPE_STRING) ? DBR_STRING :
		(nType == PV_TYPE_SHORT) ? DBR_SHORT : 0;
	return wDataType;
}

void PV_ListAll(void)
{
	int i;
	for (i = 0; i < PV_SHORT_NUM; i++)
		if (strlen(pvShort[i].szName) > 0)	
		{
			printf("%s:%s  %d", NV_Settings.szDeviceName, pvShort[i].szName, pvShort[i].sValue);
			printf("\n");
		}
	for (i = 0; i < PV_DBL_NUM; i++)
		if (strlen(pvDouble[i].szName) > 0)	
		{
			printf("%s:%s  %.2f", NV_Settings.szDeviceName, pvDouble[i].szName, pvDouble[i].dValue);
			printf("\n");
		}
	for (i = 0; i < PV_STR_NUM; i++)
		if (strlen(pvString[i].szName) > 0)	
		{
			printf("%s:%s  %s", NV_Settings.szDeviceName, pvString[i].szName, pvString[i].szValue);
			printf("\n");
		}
}

// return TRUE: if PV is in the list
//
BOOL PV_IsExist(char *szPV, int *pnType, int *pnPvNdx)
{
	int i, n;
	char *p;
	char szDev[MAX_NAME_SIZE_P1];
	//
	p = strrchr(szPV, ':');	//reverse search
	if (p == NULL) 
		return FALSE;
	n = p - szPV;
	if (n > MAX_NAME_SIZE) 
		return FALSE;		//device name too long
	memmove(szDev, szPV, n);
	szDev[n] = 0;
	//strlwr(szDev);	//make it lower case
	if (strcmp(szDev, NV_Settings.szDeviceName) != 0) 
		return FALSE;		//different device name
	p++;
	//
	for (i = 0; i < PV_SHORT_NUM; i++)
		if (strcmp(p, pvShort[i].szName) == 0) {
			*pnType = PV_TYPE_SHORT;
			*pnPvNdx = i;
			return TRUE;
		} 
	for (i = 0; i < PV_DBL_NUM; i++)
		if (strcmp(p, pvDouble[i].szName) == 0)	{
			*pnType = PV_TYPE_DOUBLE;
			*pnPvNdx = i;
			return TRUE;
		}
	for (i = 0; i < PV_STR_NUM; i++)
		if (strcmp(p, pvString[i].szName) == 0)	{
			*pnType = PV_TYPE_STRING;
			*pnPvNdx = i;
			return TRUE;
		}
	return FALSE;
}

/*
	used by file system for file header
	return 0 if no data are logged
*/
int PV_GetNameList(int nType, char *szData, int nFormat)
{
	int i;
	int nHeaderSize;
	int nLogNum = 0;

	sprintf(szData, "%s\r\n", NV_Settings.szDeviceName);
	strcat(szData, "time,");
	//
	for (i = 0; i < PV_DBL_NUM; i++) {
		if (pvDouble[i].bLogged == TRUE) {
			strcat(szData, pvDouble[i].szName);
			strcat(szData, "(D),");	//double
			nLogNum++;
		}
	}
	for (i = 0; i < PV_SHORT_NUM; i++) {
		if (pvShort[i].bLogged == TRUE) {
			strcat(szData, pvShort[i].szName);
			strcat(szData, "(S),");	//short
			nLogNum++;
		}
	}
	strcat(szData, "\r\n");
	if (nLogNum == 0)
		nHeaderSize = 0; // no log
	else if (NV_Settings.nLogFormat == FORMAT_TEXT)
		nHeaderSize = strlen(szData);
	else
		nHeaderSize = sizeof(LOG_DATA_HEADER);
	return nHeaderSize;
}

/*
	used by file system for file data
	return 0 if no data are logged
*/
int PV_GetDataList(int nType, char *szData, int nFormat)
{
	int i, nNdx;
	time_t t;
	float fData;
	short sData;
	//char szValue[30];
	int nLogNum = 0, nRet = 0, nSize;
	//
	t = time(NULL);
	t = ChangeToUTC1990(t); //change to UTC time, second since 1990/1/1 12:00:00
	if (nFormat == FORMAT_TEXT)
	{
		sprintf(szData, "%ld,", t); //save time
		for (i = 0; i < PV_DBL_NUM; i++) {
			if (pvDouble[i].bLogged == TRUE) {
				sprintf(szData, "%s%.4f,", szData, pvDouble[i].dValue);
				nLogNum++;
			}
		}
		for (i = 0; i < PV_SHORT_NUM; i++) {
			if (pvShort[i].bLogged == TRUE)	{
				sprintf(szData, "%s%d,", szData, (int)pvShort[i].sValue);
				nLogNum++;
			}
		}
		strcat(szData, "\r\n");
		nRet = (nLogNum == 0) ? 0 : strlen(szData);
	}
	else	//NV_Settings.nLogFormat == FORMAT_BINARY
	{
		memmove(szData, &t, sizeof(DWORD));
		nSize = sizeof(DWORD);
		for (i = 0, nNdx = 0; i < PV_DBL_NUM; i++) {
			if (pvDouble[i].bLogged == TRUE)	{
				fData = (float)pvDouble[i].dValue;
				memmove(&szData[nSize + nNdx * sizeof(float)], &fData, sizeof(float));
				nNdx++;
				nLogNum++;
			}
		}
		nSize = nSize + nNdx * sizeof(float);
		for (i = 0, nNdx = 0; i < PV_SHORT_NUM; i++) {
			if (pvShort[i].bLogged == TRUE)	{
				sData = pvShort[i].sValue;
				memmove(&szData[nSize + nNdx * sizeof(SHORT)], &sData, sizeof(SHORT)); 
				nNdx++;
				nLogNum++;
			}
		}
		nSize = nSize + nNdx * sizeof(short);
		nRet = (nLogNum == 0) ? 0 : nSize;
	}
	return nRet;
}
/*
// return -1 is no authorized IP found
*/
int GetAuthorizedIP(IPADDR iaV)
{
	int i, j;
	PBYTE ipa;
	PBYTE ipb = (PBYTE)&iaV;
	for (i = 0; i < MAX_AUTHORIZE_NUM; i++)
	{
		ipa = (PBYTE)&NV_Settings.iaAuthorized[i];
		for (j = 0; j < 4; j++)	{
			if ((ipa[j] == ipb[j]) || (ipa[j] == 0xFF))
				continue;
			break;
		}
		if (j != 4) //IP not all match
			continue;	//not authorized
		return i;	//authorized IP
	}
	return -1;	//unauthorized IP
}
/*
// return -1 is no authorized IP found
*/
int GetAuthorizedUser(char *szH, char *szC)
{
	int i;
	char *p1, *p2;
	char szHost[24], szClient[24];
	char szHC[65];
	//
	sprintf(szHost, "%s:", szH);
	sprintf(szClient, "%s,", szC);
	for (i = 0; i < MAX_AUTHORIZE_NUM; i++)
	{
		strncpy(szHC, NV_Settings.szAuthorizedHC[i], 63);
		szHC[63] = 0;
		strcat(szHC, ",");
		//
		p1 = strstr(szHC, szHost);
		if (p1 == NULL) 
			continue;	//host not found
		p2 = strstr(p1, szClient);
		if (p2 == NULL)
			continue;	//client not found
		#if DEBUG_FILE == 1
		iprintf("Got authorized user %d\r\n", i);
		#endif
		return i;	//SUCCESS
	}
	#if DEBUG_FILE == 1
	iprintf("Not authorized user %d\r\n", i);
	#endif
	return -1;	//unauthorized IP
}
//
int CheckAuthorization(IPADDR iaV, char *szH, char *szC, int nPvSetNdx)
{
	int nAuthorizedNdx;
	//
	nAuthorizedNdx = GetAuthorizedIP(iaV);
	if (nAuthorizedNdx >= 0) { //found IP
		pvSet[nPvSetNdx].nAuthorizedRW = NV_Settings.nAuthorizedRW[nAuthorizedNdx];
		return nAuthorizedNdx;
	}
	#if DEBUG_CA == 1
	iprintf("not authorized by IP tables\r\n");
	#endif
	if (szH == NULL || szC == NULL)
		return -1;
	nAuthorizedNdx = GetAuthorizedUser(szH, szC);
	if (nAuthorizedNdx >= 0) { //host, user pair found
		pvSet[nPvSetNdx].nAuthorizedRW = CA_PROTO_ACCESS_RIGHT_RW;
		return nAuthorizedNdx;
	}
	#if DEBUG_CA == 1
	iprintf("not authorized by host user tables\r\n");
	#endif
	return -1;
}

void TxString(char *szStr)
{
	if (nTxInterface == INTERFACE_RS232)
		printf(szStr);
	else //if (nTxInterface == FORMAT_UDP)
	{	
		{
			UDPPacket UdpTxPkt;
			UdpTxPkt.SetSourcePort(UdpClient.nDestPort);
			UdpTxPkt.SetDestinationPort(UdpClient.nSrcPort);
			UdpTxPkt.AddData((BYTE *)szStr, strlen(szStr));
			UdpTxPkt.Send(UdpClient.iaSource);
		}
	}
}

void ShowParam(void)
{
	char szStr[64];
	char szBuffer[256] = {0};
	//
	strcat(szBuffer, "Tiny CAS Controller. ");
	sprintf(szStr, "CA version %d.%d.\n", CA_PROTO_MAJ_VER, CA_PROTO_MIN_VER);
	strcat(szBuffer, szStr);
	sprintf(szStr, "Device Name: %s\n", NV_Settings.szDeviceName);
	strcat(szBuffer, szStr);
	sprintf(szStr, "FW Version %02d.%02d, %4d/%02d/%02d \n", FW_MAJOR_VER, FW_MINOR_VER, 
		FW_DATE_Y, FW_DATE_M, FW_DATE_D);
	strcat(szBuffer, szStr);
	strcat(szBuffer, "$GETDATA   get current data\n");
	strcat(szBuffer, "$GETNAME   get data name list\n");
	strcat(szBuffer, "$GETTIME   get system and UTC time\n");
	strcat(szBuffer, "$SETTIME:YYYY-MM-DD,hh:mm:ss  set system time\n");
	TxString(szBuffer);
}

void ShowUser(void)
{
	int i;
	char szRW[4][3] = {"--", "R", "RW", "RW"};
	for (i = 0; i < MAX_AUTHORIZE_NUM; i++)
	{
		iprintf("Authorized IP: ");
		ShowIP(NV_Settings.iaAuthorized[i]);
		iprintf(", %s(%d)\n", szRW[NV_Settings.nAuthorizedRW[i] & 0x03], NV_Settings.nAuthorizedRW[i]);
	}
	for (i = 0; i < MAX_AUTHORIZE_NUM; i++)
	{
		if (strlen(NV_Settings.szAuthorizedHC[i]) > 0)
			iprintf("Authorized Host:User %s\n", NV_Settings.szAuthorizedHC[i]);	//host and client
	}
	iprintf("---------------------------------------\n");
}

void ShowTitle(void)
{
	char szBuffer[32];
	//
	iprintf("---------------------------------------\n");
	iprintf("Tiny CAS Controller. ");
	iprintf("CA version %d.%d.\r\n", CA_PROTO_MAJ_VER, CA_PROTO_MIN_VER);
	IPtoString(EthernetIP, szBuffer);
	iprintf("IP address %s, ", szBuffer);
	IPtoString(NV_Settings.iaNetMask, szBuffer);
	iprintf("Netmask %s\n", szBuffer);
	// 
	iprintf("Device Name: %s\n", NV_Settings.szDeviceName);
	ShowTitleIP("NTP server: ", NV_Settings.iaNTP);
	//iprintf("Mask %d.%d.%d.%d\n", 
	//	NV_Settings.bNetMask[0], NV_Settings.bNetMask[1],
	//	NV_Settings.bNetMask[2], NV_Settings.bNetMask[3]);
	iprintf("FW Version %02d.%02d, %4d/%02d/%02d \n", FW_MAJOR_VER, FW_MINOR_VER, 
		FW_DATE_Y, FW_DATE_M, FW_DATE_D);
	iprintf("---------------------------------------\n");
}

int DecipherCommand(char *szCommand)
{
	char szDateTime[32];
	int nFormat, nInterval;
	char szFormat[3][8] = {"None", "Binary", "Text"};
	char szBuffer[128], *p;
	//
	//strupr(szCommand);		//make it upper case
	if (strncasecmp(szCommand, "GETDATA", 7) == 0)	{
		PV_GetDataList(PV_TYPE_DOUBLE, szBuffer, FORMAT_TEXT);
		TxString(szBuffer);
		return IS_A_COMMAND;
	}
	if (strncasecmp(szCommand, "GETNAME", 7) == 0)	{
		PV_GetNameList(PV_TYPE_DOUBLE, szBuffer, FORMAT_TEXT);
		TxString(szBuffer);
		return IS_A_COMMAND;
	}
	if (strncasecmp(szCommand, "GETTIME", 7) == 0)	{
		GetDateTimeString(szBuffer);
		TxString(szBuffer);
		return IS_A_COMMAND;
	}
	if (strncasecmp(szCommand, "GETNTP", 6) == 0)	{
		TimeCalibration();
		return IS_A_COMMAND;
	}
	if (strncasecmp(szCommand, "SETTIME", 7) == 0) { //YYYY-MM-DD hh:mm:ss
		SetSystemTime(&szCommand[8]);
		GetSystemTime(szDateTime, DATE_TIME_NORMAL);
		sprintf(szBuffer, "system time = %s\r\n", szDateTime);
		TxString(szBuffer);
		return IS_A_COMMAND;
	}
	if (strncasecmp(szCommand, "HELP", 4) == 0 && nTxInterface == INTERFACE_UDP) {
		ShowParam();
		return IS_A_COMMAND;
	}
	if (nTxInterface == INTERFACE_UDP)
		return NOT_A_COMMAND;
	if (strncasecmp(szCommand, "BEACON", 6) == 0) {
		bSendBeacon = (szCommand[7] == '1') ? 1 : 0; 
		sprintf(szBuffer, "beacon %d\n", bSendBeacon);
		TxString(szBuffer);
		return IS_A_COMMAND;
	}
	if (strncasecmp(szCommand, "DBL", 3) == 0)	{
		PV_ListAll();
		return IS_A_COMMAND;
	}
	if (strncasecmp(szCommand, "DIR", 3) == 0)	{
		FileProcParam.bDumpDir = TRUE;
		p = strstr(szCommand, "-d");
		FileProcParam.szDir[0] = 0;
		if (p != NULL)
			strcpy(FileProcParam.szDir, p + 2);
		return IS_A_COMMAND;
	}
	if (strncasecmp(szCommand, "FORMAT", 6) == 0) { //nTxInterface == FORMAT_TEXT
		FileProcParam.bFormat = TRUE;
		return IS_A_COMMAND;
	}
	if (strncasecmp(szCommand, "LOG", 3) == 0) {
		nFormat = atoi(&szCommand[4]);
		if (nFormat < FORMAT_NO_LOG) 
			return NOT_A_COMMAND;
		NV_Settings.nLogFormat = nFormat;
		nInterval = atoi(&szCommand[6]);
		NV_Settings.nLogInterval = (nInterval < 1) ? 1 : nInterval;
		sprintf(szBuffer, "format=%s, %d, interval=%d\r\n", szFormat[nFormat], nFormat, NV_Settings.nLogInterval);
		TxString(szBuffer);
		SaveUserParameters(&NV_Settings, sizeof(NV_SETTINGS));
		return IS_A_COMMAND;
	}
	if (strncasecmp(szCommand, "MOUNT", 5) == 0) {
		nFormat = atoi(&szCommand[6]);
		if (nFormat == 1) {
			InitExtFlash();
		}
		else if (nFormat == 0) {
			UnmountExtFlash();
		}
		return IS_A_COMMAND;
	}
	if (strncasecmp(szCommand, "REBOOT", 6) == 0) {
		ForceReboot();
		return IS_A_COMMAND;
	}
	if (strncasecmp(szCommand, "RMDIR", 4) == 0) {
		strncpy(FileProcParam.szDir, &szCommand[6], 8);
		FileProcParam.szDir[8] = 0;
		FileProcParam.bDeleteDir = TRUE;
		return IS_A_COMMAND;
	}
	if (strncasecmp(szCommand, "RESET", 5) == 0) {
		NV_Settings.dwVerifyKey = 0;
		SaveUserParameters(&NV_Settings, sizeof(NV_SETTINGS));
		ForceReboot();
		return IS_A_COMMAND;
	}
	if (strncasecmp(szCommand, "S0:", 3) == 0) {
		SerialOutput(0, &szCommand[3]);
		return IS_A_COMMAND;
	}
	if (strncasecmp(szCommand, "S1:", 3) == 0) {
		return IS_A_COMMAND;
	}
	if (strncasecmp(szCommand, "SOUT", 4) == 0) {
		NV_Settings.nSerialOutput = atoi(&szCommand[5]);
		sprintf(szBuffer, "serail output = %d\r\n", NV_Settings.nSerialOutput);
		TxString(szBuffer);
		return IS_A_COMMAND;
	}
	if (strncasecmp(szCommand, "USER", 4) == 0) {
		ShowUser();
		return IS_A_COMMAND;
	}
	if (strncasecmp(szCommand, "HELP", 4) == 0)	{
		ShowTitle();
		sprintf(szBuffer, "Log %s (%d),", szFormat[NV_Settings.nLogFormat], NV_Settings.nLogFormat);
		TxString(szBuffer);
		sprintf(szBuffer, "interval = %d s, ", NV_Settings.nLogInterval);
		TxString(szBuffer);
		sprintf(szBuffer, "Time Zone = %d min\n", NV_Settings.nTZOffset);
		TxString(szBuffer);
		TxString("---------------------------------------\n");
		TxString("BEACON <n>       enable/disable beacon\n");
		TxString("DBL              show PV\n");
		TxString("DIR [-dYYYYMM]   show all or specific dir\n");
		TxString("FORMAT           format SD card\n");
		TxString("GETDATA          get current data list\n");
		TxString("GETNAME          get data name list\n");
		TxString("GETNTP           get time from NTP server\n");
		TxString("GETTIME          get system time\n");
		TxString("LOG <n>,<i>      n:0 disable, 1 binary, 2 text\n");
		TxString("                 i: interval in sec\n");
		TxString("MOUNT <n>        1: mount, 0: unmount\n");
		TxString("REBOOT           reboot system\n");
		TxString("RESET            reset and reboot system\n");
		TxString("RMDIR <dir>      delete directory. <dir>=YYYYMM\n");
		TxString("SETTIME:YYYY-MM-DD,hh:mm:ss  set system time\n");
		TxString("SOUT <n>         serial output, 1: enable, 0: disable\n");
		TxString("---------------------------------------\n");
		return IS_A_COMMAND;
	}
	return NOT_A_COMMAND;
}
/*
	PVS function
*/
void PVS_SetMonitorMask(int nPvSetNdx, WORD wMask)
{
	/*if (wMask & DBE_ALARM != 0)
		pvSet[nPvSetNdx].wMonitorMask |= DBE_ALARM;
	if (wMask & DBE_VALUE != 0)
		pvSet[nPvSetNdx].wMonitorMask |= DBE_VALUE;
	else if (wMask & DBE_LOG != 0)
		pvSet[nPvSetNdx].wMonitorMask |= DBE_LOG;
	else*/
	pvSet[nPvSetNdx].wMonitorMask = wMask;
}	

WORD PVS_GetMonitorMask(int nPvSetNdx)
{
	return pvSet[nPvSetNdx].wMonitorMask;
}

void PVS_Clear(int nPvSetNdx)
{
	memset(&pvSet[nPvSetNdx], 0x0, sizeof(PV_SET));
	pvSet[nPvSetNdx].nPvNdx = PVS_BLANK_MARKER;
	pvSet[nPvSetNdx].nFdNdx = -1;	//can't be zero
}

void PVS_ClearFD(int nFdNdx)
{
	int i;
	for (i = 0; i < NPVS; i++)
	{
		if (pvSet[i].nFdNdx == nFdNdx)
			PVS_Clear(i);
	}
}

void PVS_SetData(int nPvSetNdx, WORD wDataType, char *szValue)
{
	int nPvNdx;
	double dValue;
	unsigned int uiV;
	//
	nPvNdx = pvSet[nPvSetNdx].nPvNdx;
	if ((wDataType == DBR_DOUBLE) || pvSet[nPvSetNdx].nType == PV_TYPE_DOUBLE)
	{
		if (wDataType == DBR_DOUBLE)
			dValue = *((double *)szValue);
		else if (szValue[0] == 'x')	//hexidecimal value
		{
			sscanf(&szValue[1], "%X", &uiV);
			dValue = (double)uiV;
		}
		else //decimal value
			dValue = atof(szValue);
		if (IOC_SetData(nPvNdx, PV_TYPE_DOUBLE, (void *)&dValue, sizeof(double)) == SUCCESS)
		{
			pvDouble[nPvNdx].dNewValue = dValue;
			pvDouble[nPvNdx].dValue = dValue; //output to analog output or digital output
		}
	}
	//if (wDataType == DBR_STRING)
	if (pvSet[nPvSetNdx].nType == PV_TYPE_STRING)
	{
		if (IOC_SetData(nPvNdx, PV_TYPE_STRING, (void *)szValue, strlen(szValue)) == SUCCESS)
		{
			strcpy(pvString[nPvNdx].szNewValue, szValue);
			strcpy(pvString[nPvNdx].szValue, szValue);
			//output to serial output, or set RTC
		}
	}
}

int PVS_GetEmptyPvSet(void)
{
	int i;
	for (i = 0; i < NPVS; i++)
		if (pvSet[i].nPvNdx == PVS_BLANK_MARKER)
			return i;
	return -1;
}

int PVS_GetPvSetNdxFromCID(DWORD dwCID)
{
	int i;
	for (i = 0; i < NPVS; i++)
		if (pvSet[i].dwCID == dwCID)
			return i;
	return -1;
}

int PVS_GetPvSetNdxFromSID(DWORD dwSID)
{
	int i;
	for (i = 0; i < NPVS; i++)
		if (pvSet[i].dwSID == dwSID)
			return i;
	return -1;
}
