/*
	utility.cpp
*/
#include "predef.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <startnet.h>
//#include <time.h>
#include <nbtime.h>
#include <basictypes.h>
#include "rtc.h"
#include "ca_proto.h"
#include "caerr.h"
#include "ffs.h"

extern NV_SETTINGS NV_Settings;
//
void ShowTitleIP(char *szName, IPADDR iaV)
{
	iprintf(szName);
	ShowIP(iaV);
	iprintf("\n");
}

WORD IntToHexWord(int nV)
{
	unsigned int uiV;
	char szBuffer[10];
	nV = nV % 10000;
	sprintf(szBuffer, "%d", nV);
	sscanf(szBuffer, "%x", &uiV);
	return (WORD)uiV;
}

int GetNearestInt(double dV)
{
	int nV;
	nV = (int)dV;
	if (dV - (double)nV > 0.5)
		return (nV + 1);
	return nV;
}

WORD ByteToWord(BYTE *pData)
{
	WORD wV;
	wV = ((WORD)pData[0]) << 8;
	wV = wV + (WORD)pData[1];
	return wV;
}
//string length must a multiple of 8
int Make8(int nN)
{
	if ((nN % 8) != 0) 
		nN = (nN / 8 + 1) * 8;
	return nN;
}
//
// generate ddd,ddd,ddd number display
//
char *Make3PNumber(long lV)
{
	char szBuffer1[24];
	static char szBuffer2[24];
	int i, j, n;
	sprintf(szBuffer1, "%ld", lV);
	n = strlen(szBuffer1);
	for (i = 0, j = 0; i < n; i++, j++)	{
		szBuffer2[j] = szBuffer1[i];
		if (((n - i) % 3 == 1) && (n - i > 1)) {
			j++;
			szBuffer2[j] = ',';
		}
	}
	szBuffer2[j] = 0;
	return szBuffer2;
}

void GetUTCTime(char *szDateTime, int nFormat)
{
	//get date time from RTC
	struct tm *p;
	time_t time_val = time(NULL);
	time_val = time_val - (long)NV_Settings.nTZOffset * 60; //change to UTC time
	p = gmtime(&time_val);  //conver to UTC time
	if (nFormat == DATE_TIME_COMPACT)
		sprintf(szDateTime, "%4d%02d%02d_%02d%02d%02d", 1900+p->tm_year, 1+p->tm_mon,
			p->tm_mday, p->tm_hour, p->tm_min, p->tm_sec); 
	else //FORMAT_TIME_NORMAL
		sprintf(szDateTime, "%4d-%02d-%02d %02d:%02d:%02d", 1900+p->tm_year, 1+p->tm_mon,
			p->tm_mday, p->tm_hour, p->tm_min, p->tm_sec); 
}
/*
	szDateTime = "YYYY-MM-DD,hh:mm:ss"  example:"2009-09-14,14:54:00"
*/
void GetSystemTime(char *szDateTime, int nFormat)
{
	//get date time from RTC
	time_t time_val = time(NULL);
	struct tm *p; 
	strcpy(szDateTime, ctime(&time_val));
	p = gmtime(&time_val);  //conver to UTC time
	if (nFormat == DATE_TIME_COMPACT)
		sprintf(szDateTime, "%4d%02d%02d_%02d%02d%02d", 1900+p->tm_year, 1+p->tm_mon,
			p->tm_mday, p->tm_hour, p->tm_min, p->tm_sec); 
	else
		sprintf(szDateTime, "%4d-%02d-%02d %02d:%02d:%02d", 1900+p->tm_year, 1+p->tm_mon,
			p->tm_mday, p->tm_hour, p->tm_min, p->tm_sec); 
}
/*
	szDateTime format = "YYYY-MM-DD,hh:mm:ss" or "YYYY-MM-DD hh:mm:ss"
	return 0: fail
	return 1: success
*/
int SetSystemTime(char *szDateTime)
{
	struct tm bts;
	time_t t;
	int nYear, nMon, nDay, nHour, nMin, nSec;
	char szBuffer[32]; //"YYYY,MM,DD,hh,mm,ss"
	//
	memmove(&szBuffer[0], &szDateTime[0], 4);
	szBuffer[4] = ',';
	memmove(&szBuffer[5], &szDateTime[5], 2);
	szBuffer[7] = ',';
	memmove(&szBuffer[8], &szDateTime[8], 2);
	szBuffer[10] = ',';
	memmove(&szBuffer[11], &szDateTime[11], 2);
	szBuffer[13] = ',';
	memmove(&szBuffer[14], &szDateTime[14], 2);
	szBuffer[16] = ',';
	memmove(&szBuffer[17], &szDateTime[17], 2);
	szBuffer[19] = 0;
	#if DEBUG_TIME == 1
	iprintf(szBuffer);
	iprintf("\r\n");
	#endif
	//
	if (sscanf(szBuffer, "%d,%d,%d,%d,%d,%d", &nYear, &nMon, &nDay, &nHour, &nMin, &nSec) != 6)
	{
		#if DEBUG_TIME == 1
		iprintf("Set time fail.\r\n");
		#endif
		return FAILURE;
	}
	
	bts.tm_isdst = 0;	//is daylight saving time?
	bts.tm_year = nYear - 1900;
	bts.tm_mon = (BYTE) nMon - 1;
	bts.tm_mday = (BYTE) nDay;
	bts.tm_hour = nHour; //0~23	
	bts.tm_min = nMin;
	bts.tm_sec = nSec;
	
	t = mktime(&bts);
	set_time(t); //set system time
	#if DEBUG_TIME == 1
	iprintf("%s\r\n", ctime(&t));
	#endif 
	//
	RTCSetTime(bts);	//set RTC chip time
	return SUCCESS;
}

//input time_val = second from 1970/1/1 12:00:00, local time
//output time = second from 1990/1/1 12:00:00, UTC time
time_t ChangeToUTC1990(time_t time_val)
{
	//time_val = second from 1970/1/1 12:00:00
	//CA time is counted from 1990/1/1 12:00:00
	//time_val is local time
	time_val = time_val - SEC_FOR_20Y - (long)NV_Settings.nTZOffset * 60; //change to UTC time
	return time_val;
}

void IPtoString(IPADDR ia, char *s)
{
	PBYTE ipb = (PBYTE)&ia;
	siprintf(s,"%d.%d.%d.%d",(int)ipb[0],(int)ipb[1],(int)ipb[2],(int)ipb[3]);
	return;
}

void PackChar(char *szBuffer, char *szPack) 
{
	char *p1, *p2;
	p2 = p1 = szBuffer;
	while (*p1 != 0) {
		if (strchr(szPack, *p1) != NULL) {	//if *p1 is in szPack 
			p1++; 
			continue; 
		}
		*p2 = *p1;
		p2++;
		p1++;
	}
	*p2 = 0;
}
